#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Copyright (C) 2009-2016 Wolfgang Rohdewald <wolfgang@rohdewald.de>

SPDX-License-Identifier: GPL-2.0-only

"""

import unittest
from typing import Optional, List, Tuple, Union, TYPE_CHECKING

from common import Debug
from wind import Wind, East, South, West, North
from player import Players
from game import PlayingGame
from hand import Hand, Score
from tile import TileTuple
from predefined import ClassicalChineseDMJL, ClassicalChineseBMJA

if TYPE_CHECKING:
    from rule import Ruleset

RULESETS = []

# test each of those rulesets twice: once with limit, once with roof off
for testRuleset in [ClassicalChineseDMJL, ClassicalChineseBMJA] * 2:
    _ = testRuleset()
    _.load()
    RULESETS.append(_)

for _ in RULESETS[2:]:
    _.roofOff = True  # type: ignore

# Do not create our test players in the data base:
Players.createIfUnknown = str  # type: ignore

# RULESETS=RULESETS[:1]
GAMES = [PlayingGame([tuple([wind, str(wind.char)]) for wind in Wind.all4], x) for x in RULESETS]  # type:ignore[misc]
PROGRAM = None


class Expected:

    """define what we expect from test"""

    def __init__(self, won:bool, points:int, doubles:int, limits:float) ->None:
        self.won = won
        self.score = Score(points, doubles, limits)
        self.ruleset:Optional['Ruleset'] = None

    def __str__(self) ->str:
        return f'Won with {self.score}' if self.won else f'Lost with {self.score}'


class Nothing(Expected):

    """we do not expect anything"""

    def __init__(self, points:int=0, doubles:int=0, limits:float=0.0) ->None:
        super().__init__(True, points, doubles, limits)


class Win(Expected):

    """we expect a winning hand"""

    def __init__(self, points:int=0, doubles:int=0, limits:float=0.0) ->None:
        super().__init__(True, points, doubles, limits)


class NoWin(Expected):

    """we expect a losing hand"""

    def __init__(self, points:int=0, doubles:int=0, limits:float=0.0) ->None:
        super().__init__(False, points, doubles, limits)


class Base(unittest.TestCase):


    """tests lots of hand examples. We might want to add comments which test should test which rule"""

    # pylint: disable=too-many-locals

    def scoreTest(self, string:str, expected:Union[Expected, List[Expected]],
        myWind:Optional[Wind]=None, roundWind:Optional[Wind]=None, totals:Optional[Tuple[int, ...]]=None) ->None:
        """execute one scoreTest test"""
        if myWind is None:
            myWind = East
        if roundWind is None:
            roundWind = East
        for idx, ruleset in enumerate(RULESETS):
            if isinstance(expected, list):
                expIdx = idx
                if expIdx >= len(expected):
                    expIdx %= len(RULESETS) // 2
                exp = expected[expIdx]
            else:
                exp = expected
            exp.ruleset = ruleset
            variants = []
            game = GAMES[idx]
            for widx, wind in enumerate(Wind.all4):
                game.players[widx].wind = wind
            game.winner = game.players[myWind]
            game.myself = game.winner
            game.point.prevailing = roundWind
            game.winner.clearCache()
            if Debug.hand:
                print('')
                print(f'starting test for {ruleset.name}')
            variant = Hand(game.winner, string)
            score = variant.score
            assert score is not None
            variants.append(variant)
            self.assertTrue(
                variant.won == isinstance(exp, Win) and score == exp.score,
                self.dumpCase(variant, exp, total=None))
# activate depending on what you are testing
#            print(string, 'expected:', expected.__str__()), variant
#            print(ruleset.name.encode('utf-8'))
#            print('\n'.join(variant.explain).encode('ut-f8'))
            if totals:
                self.assertTrue(score.total() == totals[idx], '\n'.join(
                    self.dumpCase(x, exp, totals[idx]) for x in variants))

    def callingTest(self, string:str, expected:Union[str, List[str]]) ->None:
        """test a calling hand"""
        for idx, ruleset in enumerate(RULESETS):
            game = GAMES[idx]
            game.players[0].clearCache()
            hand = Hand(game.players[0], string)
            testSays = TileTuple({x.lastTile.exposed for x in hand.callingHands}).sorted()
            if isinstance(expected, list):
                if idx >= len(expected):
                    idx %= len(RULESETS) // 2
                expIdx = idx
                if expIdx >= len(expected):
                    expIdx %= len(RULESETS) // 2
                exp = expected[expIdx]
            else:
                exp = expected
            completingTiles = TileTuple(exp)
            self.assertTrue(testSays == completingTiles,
                            f"{ruleset.name}: {string} may be completed by {completingTiles or 'None'} "
                            f"but testresult is {testSays or 'None'}")

    def dumpCase(self, hand:Hand, expected:Expected, total:Optional[int]) ->str:
        """dump test case"""
        assert self
        result = []
        result.append('')
        if not hand:
            result.append('Hand was illegal: claimed to be won but was not')
        else:
            if hand.won != isinstance(expected, Win):
                result.append(
                    f'hand.won is {hand.won}, expected {isinstance(expected, Win)}')
            result.append(hand.string)
            roofOff = ' roofOff' if hand.player.game.ruleset.roofOff else ''  #  type: ignore
            score = hand.score
            assert score is not None
            assert hand.player.game
            if score != expected.score:
                result.append(f"{hand.player.game.ruleset.name}{roofOff}: "
                              f"{'Won' if hand.won else 'Lost'} with {score} should be {expected}")
                result.append(f'hand:{hand}')
            if total is not None:
                if score.total() != total:
                    result.append(f'{id(hand.player.game.ruleset)} {hand.player.game.ruleset.name}{roofOff}: '
                                  f'total {score.total()} for {str(score)} should be {total}')
                result.append(f'hand:{hand}')
            result.extend(hand.explain())
            result.append(
                f'base={int(score.points)},doubles={int(score.doubles)},total={int(hand.total())}')
            result.append('')
        return '\n'.join(result)

# pylint: disable=useless-suppression,missing-docstring



class Partials(Base):

    """some partial hands"""

    def testMe(self) ->None:
        self.scoreTest('drdrdr fe Ldrdrdrdr', [NoWin(8, 1), NoWin(8, 2)])
        self.scoreTest('fe', [NoWin(4), NoWin(4, 1)])
        self.scoreTest('fs fw fe fn', [NoWin(16, 1), NoWin(16, 3)])
        self.scoreTest('fs ys', [NoWin(8, 1), NoWin(8, 2)], myWind=South)
        self.scoreTest('drdrdr Ldrdrdrdr', NoWin(4, 1))


class ZeroHand(Base):

    """zero hand games"""

    def testMe(self) ->None:
        self.scoreTest('c1c2c3 c7c8c9 b2b3b4 c5c5 s1s2s3 fw yw Lc1c1c2c3',
                       [Win(points=28, doubles=2), NoWin()], myWind=West, roundWind=North)
        self.scoreTest('c1c2c3 c7c8c9 b2b3b4 drdr s1s2s3 fw yw Lc1c1c2c3',
                       [Win(points=30, doubles=1), NoWin()], myWind=West, roundWind=North)


class FalseColorGame(Base):

    """false color games"""

    def testMe(self) ->None:
        self.scoreTest(
            'c1c1c1 c7c7c7 c2c3c4 c5c5 c6c6c6 Lc5c5c5', [Win(32, 3), Win(28)])
        self.scoreTest('c1c2c3 wewewe drdrdr dbdb DgDgDg Ldbdbdb', [
            Win(44, 4), Win(38, 2)], myWind=West, roundWind=North)
        self.scoreTest(
            's1s1s1 wewewe c2c3c4 c5c5 c6c6c6 Lc5c5c5',
            [Win(34),
             Win(30)],
            myWind=West, roundWind=North)
        self.scoreTest(
            'RB1B1B1B1B2B3B4B5B6B7B8B9DrDr fe ys LDrDrDr', [Win(46, 2), NoWin()])
        self.scoreTest(
            'b1B1B1b1 RB2B3B4B5B6B7B8B8B8 DrDr fe ys LDrDrDr', [Win(74, 2), NoWin()])
        self.scoreTest('b1B1B1b1 RB2B2B2B5B6B7B8B8B8 DrDr fe ys LDrDrDr', [
            Win(78, 3), Win(72, 1)], myWind=West)


class WrigglingSnake(Base):

    """the wriggling snake as in BMJA"""

    def testMe(self) ->None:
        self.scoreTest(
            'c1c1 c2c3c4 c5c6c7 RC8C9WeWwWsWn Lc1c1c1', [NoWin(), NoWin()])
        self.scoreTest(
            'c2c3c4 c5c6c7 RC1C1C8C9WeWwWsWn LC1C1C1', [NoWin(), NoWin()])
        self.scoreTest(
            'c2c3c4 c5c6c7 RC1C1C8C9WeWwWsWn LWnWn', [NoWin(), NoWin()])
        self.scoreTest(
            'RC1C1C2C3C4C5C6C7C8C9WeWwWsWn LC1C1',
            [NoWin(),
             Win(limits=1)])
        self.callingTest('RS1S3WwS6WsS3S3WnWeS5 s7s8s9 fs', '')
        self.callingTest('RS1S2WwS6WsS3S3WnWeS5S7S8S9 fs', '')
        self.callingTest('RS1S2WwS6WsS3S4WnWeS5S7S8S9 fs', ['', 's1'])
        self.callingTest('RS1S2WwS6WsS3S4WnWeS1S7S8S9 fs', ['', 's5'])


class SquirmingSnake(Base):

    """the winding snake"""

    def testMe(self) ->None:
        self.scoreTest(
            'c1c1c1 c3c4c5 c9c9c9 c6c7c8 RC2C2 Lc1c1c1c1',
            [Win(limits=1),
             NoWin()])
        self.scoreTest('c1c1c1 c4c5c6 c9c9c9 c6c7c8 RC2C2 Lc1c1c1c1', [
            Win(points=28, doubles=3), NoWin()])
        self.scoreTest(
            'c1c1c1 c3c4c5 c9c9c9 c6c7c8 RS2S2 Lc1c1c1c1',
            [Win(points=28),
             NoWin()])
        self.scoreTest(
            's1s1s1 s2s3s4 s9s9s9 s6s7s8 RS5S5 Ls1s1s1s1',
            [Win(limits=1),
             NoWin()])
        self.scoreTest(
            'b1b1b1 c3c4c5 c6c7c8 c9c9c9 RC2C2 Lc3c3c4c5',
            [Win(points=28),
             NoWin()])
        self.scoreTest(
            'b1b1b1 c3c4c5 c6c7c8 c9c9c9 RC2C2 Lc4c3c4c5',
            [Win(points=30),
             NoWin()])
        self.scoreTest('RC1C1C1C2C3C4C5C6C7C8C9C9C9C5 LC5', Win(limits=1))
        self.scoreTest(
            'RC1C1C1C3C4C5C6C7C8C9C9C9C5 LC3',
            [NoWin(16),
             NoWin(limits=0.4)])
        self.scoreTest(
            'RC1C1C2C3C4C5C6C7C8C9C9C9C5 LC3',
            [NoWin(8),
             NoWin(limits=0.4)])


class Purity(Base):

    """Purity BMJA"""

    def testMe(self) ->None:
        self.scoreTest(
            'b1b1b1b1 RB2B3B4B5B6B7B8B8B2B2B2 fe fs fn fw LB3B2B3B4',
            [Win(points=60, doubles=4), NoWin()])
        self.scoreTest(
            'b1b1b1 RB3B3B3B6B6B6B8B8B2B2B2 fe fs fn fw LB3', [Win(54, 6), Win(54, 7)])
        self.scoreTest(
            'b1b1b1 RB3B3B3B6B6B8B8B2B2B2 fe fs fn fw LB3', [NoWin(28, 1), NoWin(28, 6)])
        self.scoreTest(
            'c1C1C1c1 c3c3c3 c8c8c8 RC4C5C6C7C7 fs fw ys yw Lc8c8c8c8',
            [Win(72, 4), Win(72, 2)], myWind=West)


class TrueColorGame(Base):

    """true color games"""

    def testMe(self) ->None:
        self.scoreTest(
            'b1b1b1b1 RB2B3B4B5B6B7B8B8B2B2B2 fe fs fn fw LB3B2B3B4',
            [Win(points=60, doubles=4), NoWin()])
        self.callingTest(
            'RB1B2B3B4B5B5B6B6B7B7B8B8B8 LB1', ['b1b3b4b6b7b9', ''])
        self.scoreTest(
            'b1b1b1B1 RB2B3B4B5B6B7B8B8B2B2B2 fe fs fn fw LB3B2B3B4',
            [Win(limits=1),
             NoWin()])


class OnlyConcealedMelds(Base):

    """only concealed melds"""

    def testMe(self) ->None:
        self.scoreTest(
            'RB1B1B1B1B2B3B4B5B6B7B8B9DrDr fe ys LDrDrDr', [Win(46, 2), NoWin()])
        self.scoreTest('RB1B1B1B2B2B2B4B4B4B7B8B9DrDr fe ys LDrDrDr', [
            Win(54, 3), Win(48, 1)], myWind=West)
        self.scoreTest(
            'b1B1B1b1 RB2B3B4B5B6B7B8B8B8DrDr fe ys LDrDrDr', [Win(74, 2), NoWin()])
        self.scoreTest('b1B1B1b1 RB2B2B2B5B6B7B8B8B8DrDr fe ys LDrDrDr', [
            Win(78, 3), Win(72, 1)], myWind=West)


class LimitHands(Base):

    """various limit hands"""

    def testMe(self) ->None:
        self.scoreTest(
            'c1c1c1 c9c9 b9b9b9b9 s1s1s1 s9s9s9 Lc1c1c1c1',
            Win(limits=1))
        self.scoreTest(
            'c1c1c1c1 drdr wewewewe c3c3c3C3 s1S1S1s1 Ldrdrdr',
            Win(limits=1))
        self.scoreTest(
            'drdr c1c1c1c1 wewewewe c3c3c3C3 s1S1S1s1 Ldrdrdr',
            Win(limits=1))
        self.scoreTest(
            'c1c1c1c1 wewewewe c3c3c3C3 s1S1S1s1 drdr Ldrdrdr',
            Win(limits=1))


class AllGreen(Base):

    """the green hand"""

    def testMe(self) ->None:
        self.scoreTest(
            'c1c1c1 c7c7c7 c2c3c4 c5c5 c6c6c6 Lc5c5c5', [Win(32, 3), Win(28)])
        self.scoreTest(
            'b2b2b2b2 RDgDgDg b6b6b6 b4b4b4 b8b8 Lb6b6b6b6',
            Win(limits=1))
        self.scoreTest(
            'b2b2b2b2 RDgDg b6b6b6 b4b4b4 b8b8 Lb6b6b6b6',
            [NoWin(14),
             NoWin(limits=0.4)])
        self.scoreTest(
            'b1b1b1b1 RDgDgDg b6b6b6 b4b4b4 b8b8 Lb6b6b6b6', [Win(48, 3), Win(48, 2)])


class NineGates(Base):

    """the nine gates"""
        # DMJL allows 1..9 as last tile, BMJA allows only 2..8

    def testMe(self) ->None:
        self.scoreTest('RC1C1C1C2C3C4C5C6C7C8C9C9C9C5 LC5', Win(limits=1))
        self.scoreTest(
            'RC1C1C1C2C3C4C5C6C7C8C9C9C9C5 LC6',
            [Win(limits=1),
             Win(limits=1)])
        self.scoreTest(
            'RC1C1C1C2C3C4C5C6C7C8C9C9C9C9 LC9',
            [Win(limits=1),
             NoWin()])
        self.scoreTest(
            'RC1C1C1C2C3C4C5C6C7C8C9C9C9C9 LC9',
            [Win(limits=1),
             NoWin()])
        self.scoreTest('RC1C1C1C2C3C4C5C6C7C8C9C9C9C5 LC2', Win(limits=1))
        self.scoreTest(
            'RC1C1C1C2C3C4C5C6C7C8C9C9C9C9 LC1',
            [Win(limits=1),
             NoWin()])
        self.scoreTest('RC1C1C2C3C4C5C6C7C8C9C9C9C9 LC1', NoWin(8))
        self.scoreTest(
            'RC1C1C2C3C4C5C6C7C8C8C9C9C9 LC1',
            [NoWin(8),
             NoWin(limits=0.4)])


class Manual(Base):

    """some manual rules for manual scoring"""
        # this should actually never happen but anyway we want to be sure that no rule
        # fires on this

    def testMe(self) ->None:
        self.scoreTest('', NoWin())
        self.scoreTest('', NoWin())


class ThirteenOrphans(Base):

    """The 13 orphans"""

    def testMe(self) ->None:
        self.scoreTest('RC1C9B9B1S1S9WeDgWsWnWwDbDrS1 LDgDg', Win(limits=1))
        self.scoreTest('ww RC1C9B9B1S1S9WeDgWsWnDbDrS8 Lwwww', NoWin())
        self.scoreTest('ww RC1C9B9B1S1S9WeDgWsWnDbDrS9 Lwwww', Win(limits=1))
        self.scoreTest('RC1C9B9B1S1S9S9WeDgWsWnWwDbDr LDrDr', Win(limits=1))
        self.scoreTest('dr RC1C9B9B1S1S9S9WeDgWsWnWwDb Ldrdr', Win(limits=1))
        self.scoreTest(
            'RC1C9B9B1S1S9S9WeDgWnWwDbDr LDb',
            [NoWin(),
             NoWin(limits=0.4)])
        self.callingTest('Dg Dg Dr We Ws Ww Wn Wn RB1B9C1S1S9 LWe', '')
        self.callingTest('Db Dg Dr We Ws Ww Wn B7 RB1B9C1S1S9 LWe', '')
        self.callingTest('RDbDgDrWeWsWwWnWnB1B9C1S1S9 LWn', 'c9')
        self.callingTest('RDbDgDrWsWwWnWnB1B9C1S1S9C9 LDg', 'we')


class SimpleNonWinningCases(Base):

    """normal hands"""

    def testMe(self) ->None:
        self.scoreTest('s2s2s2 s2s3s4 RB1B1B1B1 c9c9c9C9 Ls2s2s3s4', NoWin(26))


class FourBlessingsOverTheDoor(Base):

    """lots of winds"""

    def testMe(self) ->None:
        self.scoreTest(
            'b1b1 wewewe wswsws WnWnWn wwwwwwww Lb1b1b1',
            Win(limits=1))
        self.scoreTest(
            'RDgDg wewewe wswsws WnWnWn wwwwwwww LDgDgDg',
            Win(limits=1))
        self.scoreTest(
            'wewewe wswsws WnWnWn wwwwwwww DrDr LDrDrDr',
            Win(limits=1))
        self.scoreTest(
            'wewewe wswsws WnWnWn wwwwwwww DrDr LDrDrDr',
            Win(limits=1))
        self.scoreTest(
            'wewewe wswsws WnWnWn wwwwwwww DrDr mz LDrDrDr',
            Win(limits=1))
        self.scoreTest(
            'wewewe wswsws RWnWnWnDr wwwwwwww', [NoWin(32, 4), NoWin(limits=0.4)])


class AllHonors(Base):

    """only honors"""

    def testMe(self) ->None:
        self.scoreTest(
            'drdrdr wewe wswsws wnwnwn dbdbdb mz Ldrdrdrdr',
            Win(limits=1))
        self.scoreTest(
            'wewewe wswsws RWnWnWnB1 wwwwwwww LB1', [NoWin(32, 4), NoWin(limits=0.4)])
        # this one is limits=0.4 = 400 points, but points 512 are higher:
        self.scoreTest(
            'wewewe drdrdr RDrDrDrDb wwwwwwww LDb', [NoWin(32, 4), NoWin(32, 4)])
        self.scoreTest('wewe wswsws RWnWnWn wwwwwwww b1b1 Lwewewe', [
            NoWin(30, 2), NoWin(30, 1)], myWind=North)


class BuriedTreasure(Base):

    """buried treasure, CC BMJA"""

    def testMe(self) ->None:
        self.scoreTest('RWeWeWeC3C3C3S3S3 c4c4c4C4 b8B8B8b8 me LWeWeWeWe',
                       [Win(limits=1), Win(58, 4)])
        self.scoreTest('RWeWeWeC3C3C3S3S3C4C4C4B8B8B8 me LWeWeWeWe',
                       [Win(limits=1), Win(42, 4)])
        self.scoreTest('RWeWeWeC3C3C3C5C5C4C4C4C8C8C8 me LWeWeWeWe',
                       [Win(limits=1), Win(limits=1)])
        self.scoreTest('RWeWeC3C3C3C5C5C4C4C4C8C8C8 me LWe',
                       [NoWin(16, 1), NoWin(limits=0.4)])
        self.scoreTest('RWeWeWeC3C3C3C5C5C4C4C4C7C8C9 me LWeWeWeWe',
                       [Win(38, 6), Win(38, 3)])


class HiddenTreasure(Base):

    """hidden treasure, CC DMJL"""

    def testMe(self) ->None:
        self.scoreTest('RWeWeWeC3C3C3S3S3 c4c4c4C4 b8B8B8b8 me LWeWeWeWe',
                       [Win(limits=1), Win(58, 4)])
        self.scoreTest('RWeWeWeC3C3C3S3S3 c4c4c4C4 b8B8B8b8 LC3C3C3C3',
                       [Win(limits=1), Win(58, 4)])
        self.scoreTest('RWeWeWeC3C3C3 c4c4c4C4 b8B8B8b8 s3s3 Ls3s3s3',
                       [Win(60, 4), Win(56, 3)])


class FourfoldPlenty(Base):

    """4 kongs"""

    def testMe(self) ->None:
        self.scoreTest('RB3B3B3C1C1C1 b1b1b1 s3s4s5 wewe LB3B3B3B3', Win(42))
        self.scoreTest(
            'b3B3B3b3 c1C1C1c1 b1b1b1b1 s3s3s3s3 wewe Lwewewe',
            Win(limits=1))
        self.scoreTest(
            'b3B3B3b3 c1C1C1c1 b1b1b1b1 s3s3s3s3 WeWe LWeWeWe',
            Win(limits=1))
        self.scoreTest(
            'b3b3 c1C1C1c1 b1b1b1b1 s3s3s3s3 wewewewe Lb3b3b3',
            Win(limits=1))
        self.scoreTest(
            'b3b3b3b3 c1c1 b1b1b1b1 s3s3s3s3 wewewewe Lc1c1c1',
            Win(limits=1))
        self.scoreTest(
            'b3b3b3b3 RC1 b1b1b1b1 s3s3s3s3 wewewewe', [NoWin(48, 2), NoWin(limits=0.4)])


class PlumBlossom(Base):

    """Gathering the plum blossom from the roof"""

    def testMe(self) ->None:
        self.scoreTest(
            's2s2s2 RS5S5S5B1B1B1B2B2 c9C9C9c9 me LS5S5S5S5',
            Win(limits=1))
        self.scoreTest(
            's2s2s2 RS5S5S5B1B1B1B2B2 c9C9C9c9 me Ls2s2s2s2', [Win(66, 3), Win(66, 1)])
        self.scoreTest(
            's2s2s2 RS5S5S5B1B1B1B2B2 c9C9C9c9 LS5S5S5S5', [Win(68, 2), Win(68, 1)])


class PluckingMoon(Base):

    """plucking the moon from the bottom of the sea"""

    def testMe(self) ->None:
        self.scoreTest(
            's2s2s2 RS1S1S1B1B1B1B2B2 c9C9C9c9 mz LS1S1S1S1',
            Win(limits=1))
        self.scoreTest(
            's2s2s2 RS1S1S1B1B1B1B2B2 c9C9C9c9 mz Ls2s2s2s2', [Win(70, 3), Win(70, 2)])
        self.scoreTest(
            's2s2s2 RS1S1S1B1B1B1B2B2 c9C9C9c9 LS1S1S1S1', [Win(72, 2), Win(72, 1)])


class ScratchingPole(Base):

    """scratch a carrying pole"""

    def testMe(self) ->None:
        self.scoreTest(
            'b2b3b4 RS1S1S1B1B1B1B4B4 c9C9C9c9 mk Lb2b2b3b4',
            Win(limits=1))
        self.scoreTest(
            'b2b2b2 RS1S1S1B1B1B1B4B4 c9C9C9c9 LS1S1S1S1', [Win(72, 2), Win(72, 1)])
        self.scoreTest(
            'b2b2b2 RS1S1S1B1B1B1B4B4 c9C9C9c9 Lb2b2b2b2', [Win(70, 2), Win(70, 1)])


class ThreeGreatScholars(Base):

    """three concealed pungs"""

    def testMe(self) ->None:
        self.scoreTest('dgdgdg RDrDrDrDbDbDb s4s4s4 c5c5', Win(limits=1))
        self.scoreTest(
            'dgdgdg RDrDrDrDbDb s4s4s4 c5c5', [NoWin(16, 3), NoWin(limits=0.4)])
        self.scoreTest(
            'dgdgdg RDrDrDrDbDbDb s4s5s6 c5c5', [Win(limits=1), Win(40, 3)])
        # calling for Three Great Scholars:
        self.scoreTest(
            's2s2 RDgDgDgDbDbDbDrDrDr b2b2b2b2 Ls2s2s2',
            Win(limits=1))
        self.scoreTest(
            'RDgDgDgDbDbDbDrDrDrS2 b2b2b2b2 LDbDbDbDb', [NoWin(32, 6), NoWin(limits=0.4)])
        # 40, 5 is more than 0.4 limits:
        self.scoreTest(
            'RS2DgDgDgDbDbDbDrDrDr wewewewe LDbDbDbDb', [NoWin(40, 8), NoWin(40, 5)])


class ThreeConcealedPungs(Base):

    """three concealed pungs"""

    def testMe(self) ->None:
        self.scoreTest(
            'RB2B2B2S1S1S1B1B1B1B4B4 c9c9c9c9 LB2B2B2B2', [Win(58, 2), Win(58, 1)])
        self.scoreTest(
            'RB2B2B2S1S1S1B4B4 b1b1b1 c9c9c9c9 LB2B2B2B2', Win(54, 1))
        self.scoreTest(
            'RB2B2B2S1S1S1B4B4 b1b1b1 c9c9c9C9 LB2B2B2B2', [Win(54, 2), Win(54, 1)])


class Rest(Base):

    """various tests"""

    def testMe(self) ->None:
        self.scoreTest('s1s1s1s1 s2s2s2 wewe RS3S3S3 s4s4s4 Ls2s2s2s2',
                       [Win(44, 2), Win(44, 1)], myWind=South, roundWind=West)
        self.scoreTest('s1s1s1s1 s2s2s2 RWeWeS3S3S3 s4s4s4 me LS3S3S3S3',
                       [Win(46, 3), Win(46, 1)], myWind=South, roundWind=West)
        self.scoreTest(
            'b3B3B3b3 RDbDbDbDrDrDr wewewewe s2s2 Ls2s2s2', [Win(72, 6), Win(68, 5)])
        self.scoreTest('s1s2s3 s1s2s3 b3b3b3 b4b4b4 RB5 fn yn LB5', [
            NoWin(12, 1), NoWin(12, 2)], myWind=North)
        self.scoreTest(
            'b3b3b3b3 RDbDbDb drdrdr weWeWewe s2s2 Ls2s2s2', [Win(76, 5), Win(72, 5)])
        self.scoreTest('s2s2s2 s2s3s4 RB1B1B1B1 c9C9C9c9 Ls2s2s3s4', NoWin(42))
        self.scoreTest(
            's2s2s2 RDgDgDbDbDbDrDrDr b2b2b2b2 Ls2s2s2s2', [Win(48, 4), Win(48, 3)])
        self.scoreTest(
            's1s1s1s1 s2s2s2 s3s3s3 s4s4s4 s5s5 mw Ls3s3s3s3', Win(42, 4))
        self.scoreTest(
            'RB2C1B2C1B2C1WeWeS4WeS4WeS6 LC1', [NoWin(20, 3), NoWin(20, 2)])
        self.scoreTest('b6b6b6 RB1B1B2B2B3B3B7S7C7B8 LB3', NoWin(2))
        self.scoreTest('RB1B1B1B1B2B3B4B5B6B7B8B9DrDr fe fs fn fw LDrDrDr',
                       [Win(54, 3), NoWin()])
        self.scoreTest('RB1B1B1B2B2B2B5B5B5B7B8B9DrDr fe fs fn fw LDrDrDr',
                       [Win(62, 4), Win(56, 4)])
        self.scoreTest('RB1B1B1B1B2B3B4B5B6B7B8B9DrDr fe fs fn fw me LDrDrDr',
                       [Win(54, 4), NoWin()])
        self.scoreTest('RB1B1B1B1B2B3B4B4B4B7B7B7DrDr fe fs fn fw mz LDrDrDr',
                       [Win(62, 5), Win(56, 5)])
        self.scoreTest('RB1B1B1B1B2B3B4B4B4B7B7B7DrDr fe fs fn fw mZ LDrDrDr',
                       [Win(62, 5), Win(56, 5)])
        self.scoreTest('drdr RB1B1B1B1B2B3B4B5B6B7B8B9 fe fs fn fw mZ Ldrdrdr',
                       [Win(52, 3), NoWin()])
        self.scoreTest(
            'RB1B1B1B1B2B3B4B5B6B7B8B8B2B2 fe fs fn fw LB4',
            NoWin())
        self.scoreTest('RB1B1B1B1B2B3B4B5B6B8B8B2B2 fe fs fn fw LB4',
                       [NoWin(28, 1), NoWin(28, 3)])
        self.scoreTest('wewe wswsws RWnWnWn wwwwwwww b1b1b1 mz Lb1b1b1b1',
                       [Win(54, 6), Win(54, 4)], myWind=North)
        self.scoreTest('wswsws RWeWeWnWnWnB1B1B1 wwwwwwww mz LB1B1B1B1',
                       [Win(60, 6), Win(60, 4)], myWind=North)
        self.scoreTest(
            'RB2B2 b4b4b4 b5b6b7 b7b8b9 c1c1c1 md Lb7b7b8b9', [Win(28), NoWin()])
        self.scoreTest(
            'RB8B8 s4s4s4 b1b2b3 b4b5b6 c1c1c1 md Lb3b1b2b3', [Win(28), NoWin()])
        self.scoreTest(
            'RB2B2 s4s4s4 b1b2b3 b4b5b6 c1c1c1 md Lb3b1b2b3', [Win(26), NoWin()])
        self.scoreTest(
            'RB2B2 s4s4s4 b1b2b3 b4b5b6 c1c1c1 md Lb3b1b2b3', [Win(26), NoWin()])


class TwofoldFortune(Base):

    """Twofold fortune"""

    def testMe(self) ->None:
        self.scoreTest(
            'b1B1B1b1 RB2B3B4B5B6B7 b8b8b8b8 b5b5 fe fs fn fw m.t LB4',
            [Win(limits=1),
             NoWin()])
        self.scoreTest(
            'b1B1B1b1 RB2B3B4B6B6B6 b8b8b8b8 b5b5 fe fs fn fw m.t LB4',
            Win(limits=1))


class OriginalCall(Base):

    """original call"""

    def testMe(self) ->None:
        # in DMJL, b4 would also win:
        self.scoreTest('s1s1s1 s1s2s3 RB6B6B6B8B8B8B5B5 fn yn m.a LB5B5B5',
                       [Win(44, 2), Win(42, 3)], myWind=North)
        self.scoreTest('s1s1s1 s1s2s3 RB6B6B6B8B8B8B5 fn yn m.a LB5',
                       [NoWin(20, 1), NoWin(20, 2)], myWind=North)


class RobbingKong(Base):

    """robbing the kong"""

    def testMe(self) ->None:
        # this hand is only possible if the player declared a hidden chow.
        # is that legal?
        self.scoreTest('s1s2s3 s1s2s3 RB6B6B7B7B8B8B5 fn yn m.a LB5',
                       [NoWin(8, 1), NoWin(8, 2)], myWind=North)
        self.scoreTest('s1s2s3 s2s3s4 RB6B6B7B7B8B8B5B5 fn yn mka Ls1s1s2s3',
                       [Win(28, 4), NoWin()], myWind=North)
        self.scoreTest('s4s5s6 RS1S2S3B6B6B7B7B8B8B5B5 fn yn m.a LS1S1S2S3',
                       [Win(30, 3), NoWin()], myWind=North)
        self.scoreTest('s4s5s6 RS1S2S3B6B6B7B7B8B8 fn yn m.a Ls4s4s5s6',
                       [NoWin(8, 1), NoWin(8, 2)], myWind=North)


class Blessing(Base):

    """blessing of heaven or earth"""

    def testMe(self) ->None:
        self.scoreTest('S4S5S6 RS1S2S3B6B6B7B7B8B8B5B5 fn yn m1',
                       [Win(limits=1), NoWin()])
        self.scoreTest('S4S5S6 RS1S2S3B6B6B7B7B8B8B5B5 fn yn m1',
                       [Win(limits=1), NoWin()])
        self.scoreTest('S4S5S6 RS1S1S1B6B6B6B8B8B8B5B5 fn yn m1',
                       Win(limits=1))
        self.scoreTest('s4s5s6 RS1S2S3B6B6B7B7B8B8 b5b5 fn yn m1 Ls4s4s5s6',
                       [Win(limits=1), NoWin()], myWind=West)


class Terminals(Base):

    """only terminals"""

    def testMe(self) ->None:
        # must disallow chows:
        self.scoreTest(
            'b1b1 c1c2c3 c1c2c3 c1c2c3 c1c2c3 Lb1b1b1', [Win(26, 1), NoWin()])
        self.scoreTest(
            'b1b1 c1c2c3 c9c9c9 s1s1s1 s9s9s9 Lb1b1b1', [Win(38), Win(32)])
        self.scoreTest(
            'b1b1 c1c1c1 c9c9c9 s1s1s1 s9s9s9 Lb1b1b1',
            Win(limits=1))
        self.scoreTest(
            'RB1 c1c1c1 c9c9c9 s1s1s1 s9s9s9',
            [NoWin(16),
             NoWin(limits=0.4)])


class LongHand(Base):

    """long hand"""

    def testMe(self) ->None:
        self.scoreTest('s1s2s3 s1s2s3 b3b3b3 b4b4b4 RB5B6 fn yn LB5', NoWin())
        self.scoreTest('RB2C1B2C1B2C1WeWeS4WeS4WeS6S5 LS5', NoWin())
        self.scoreTest('RB2C1B2C1B2C1WeWeS4WeS4WeS6S5S5 LS5', NoWin())
        self.scoreTest('RB2C1B2C1B2C1WeWeS4WeS4WeS6S5S5 LS5', NoWin())
        self.scoreTest('RWsWsWsWsWnS6C1C1WeWeWeS4S4S5S5 LS5', NoWin())


class MJ(Base):

    """test winner hands.
        Are the hidden melds grouped correctly?"""

    def testMe(self) ->None:
        self.scoreTest(
            'RB1B1B1B2B2B2B3B4 wnwnwn wewewe Lwnwnwnwn', [Win(36, 3), Win(36, 2)])
        self.scoreTest(
            'RB1B1B1B2B2B2B3B3B3S1S1 c3c4c5 Lc3c3c4c5', [Win(36, 1), Win(36)])
        self.scoreTest(
            'RB1B1B1B2B2B2B3B3S1S2S3 c3c4c5 Lc3c3c4c5', [Win(32), NoWin()])
        self.scoreTest(
            'c1C1C1c1 b5B5B5b5 c2C2C2c2 c3C3C3c3 RC4B6 fs fw fn fe LC4',
            NoWin())
        self.scoreTest(
            'wewewe wswsws wnwnwnWn RWwWwWwC6 LC6', [NoWin(32, 4), NoWin(limits=0.4)])
        self.scoreTest('wewewe wswsws wnwnwnWn RWwWwWwC3B6 LC3', NoWin())
        self.scoreTest('wewewe wswsws wnwnwnWn RWwWwWwC3C3 LC3', Win(limits=1))


class LastIsOnlyPossible(Base):

    """tests for determining if this was the only possible last tile"""

    def testMe(self) ->None:
        self.scoreTest(
            's2s3s4 DrDr S1S2S3 S6S7S8 B5B6B7 fw yw fs md Ls2s2s3s4', [Win(34), NoWin(0)])
        self.scoreTest('b3B3B3b3 wewewewe s2s2 RDbDbDbDrDrDr Ls2s2s2',
                       [Win(72, 6), Win(68, 5)], totals=(500, 1000, 4608, 2176))
        self.scoreTest(
            'b3B3B3b3 wewewe RDbDbDbS1S1S1S2S2 LS2S2S2', [Win(60, 5), Win(58, 4)])
        self.scoreTest(
            'b3B3B3b3 wewewe RDbDbDbS1S1S1S3S3 LS3S3S3', [Win(60, 5), Win(58, 4)])
        self.scoreTest(
            'b3B3B3b3 wewewe RDbDbDbS1S1S1S4S4 LS4S4S4', [Win(62, 5), Win(58, 4)])
        self.scoreTest(
            'b3B3B3b3 wewewe RDbDbDbS1S1S1S3S3 LS1S1S1S1', [Win(58, 5), Win(58, 4)])
        self.scoreTest(
            's9s9s9 s8s8s8 RDgDgS1S2S3S3S4S5 LS3 fe', [Win(34, 1), NoWin()])
        self.scoreTest(
            's9s9s9 s8s8s8 RDgDgS1S2S3S4S4S4 LS3 fe', [Win(40, 1), Win(38, 1)])
        self.scoreTest(
            's9s9s9 s8s8s8 RDgDgS1S2S3S4S5S6 LS6 fe', [Win(34, 1), NoWin()])
        self.scoreTest(
            's9s9s9 s8s8s8 RDgDgS1S2S3S7S8S9 LS7 fe', [Win(36, 1), NoWin()])


class TripleKnitting(Base):

    """triple knitting BMJA"""

    def testMe(self) ->None:
        self.scoreTest(
            'RS2B2C2S4B4C4S6B6C6S7B7C7 s8b8 Ls8s8b8',
            [Nothing(),
             Win(limits=0.5)])
        self.scoreTest(
            'RB2C2S4B4C4S6B6C6S7B7C7S8B8 LS8',
            [NoWin(),
             NoWin(limits=0.2)])
        self.scoreTest(
            'RS2B2C2S4B4C4S6B6C6S7B7C7S8 LS8',
            [NoWin(),
             NoWin(limits=0.2)])
        self.scoreTest(
            'RS2B2C2S7B7C7S4B4C4S6B6C6 s8b8 Ls8s8b8',
            [NoWin(),
             Win(limits=0.5)])
        self.scoreTest('RS2B2C2S6B6C6S7B7C7 s4b4c4 s8b8 Ls8s8b8', NoWin())
        self.scoreTest(
            'RS2B2C2S4B4C4S6B6C6S4B4C4 s8c8 Ls8s8c8',
            [NoWin(),
             Win(limits=0.5)])
        self.scoreTest(
            'RS2B2C2S3B3C3S4B4C4S4B4C4 s8c8 Ls8s8c8',
            [NoWin(),
             Win(limits=0.5)])
        self.scoreTest('RB2C2B3C3B4C4S4B4C4 s2s3s4 s8c8 Ls8s8c8', NoWin())
        self.scoreTest('RB5S7B7B3B9S7S6C9C3B7S3 c5c6c7 mw LS3', NoWin())
        self.scoreTest('RC1C2C5C8C9S2S4S9B4B5B6B6B8S1', NoWin())
        self.scoreTest('RC1C2C5C8C9S2S4S9B4B5B6B6B8', NoWin())
        self.scoreTest(
            'RS2B2C2S4B4C4S6B6C6S4B4C4S8',
            [NoWin(),
             NoWin(limits=0.2)])
        self.scoreTest('RS1B1C1S2B2C2S5B5S8B8C8B9C7C9 LB2', NoWin())
        self.callingTest('RS2B2C2S4B4C4S6B6C6S7B7C7S8 LS8', ['', 'b8c8'])
        self.callingTest('RS2B2C2S4B4C4S6B6C6B7C7S8C8 LC7', ['', 's7b8'])


class Knitting(Base):

    """knitting BMJA"""

    def testMe(self) ->None:
        self.scoreTest(
            'RS2B2S3B3S4B4S5B5S6B6S7B7 s9b9 Ls9s9b9',
            [NoWin(),
             Win(limits=0.5)])
        self.scoreTest(
            'RS2B2S3B3B4S5B5S6B6S7B7 S9B9 LS9',
            [NoWin(),
             NoWin(limits=0.2)])
        self.scoreTest('RS2B2S3S5B5B3S4B4S6B6S7B7 s9c9 Ls9s9c9', NoWin())
        self.scoreTest('RS2B3S3B3S4B4S5B5S6B6S7B7 s9b9 Ls9s9b9', NoWin())
        self.scoreTest(
            'RS2B2S2B2S4B4S5B5S6B6S7B7 s9b9 Ls9s9b9',
            [NoWin(),
             Win(limits=0.5)])
        self.scoreTest(
            'RS2S3S4S5S6S7S9B2B3B4B5B6B7B9 LB9',
            [NoWin(),
             Win(limits=0.5)])
        self.scoreTest(
            'RS3S4S5S6S7S9B2B3B4B5B6B7B9 LB9',
            [NoWin(),
             NoWin(limits=0.2)])
        self.scoreTest('RS3S4S9S9B1B1B2B3B4B8B8 s2s2s2 fs mw LS4', NoWin())
        self.scoreTest('RS3S9S9B1B1B2B3B8B8S2S2S2 s4b4 fs mw Ls4', NoWin())
        self.callingTest('RS2B2S3B3S4B4S5B5S6B6S7B7S9 LB7', ['s9', 'b9'])


class AllPairHonors(Base):

    """all pairs honors BMJA"""

    def testMe(self) ->None:
        self.scoreTest(
            'RWeWeS1S1B9B9DgDgDrDrWsWsWwWw LS1S1S1',
            [NoWin(),
             Win(limits=0.5)])
        self.scoreTest(
            'RWeWeS1S1B9B9DgDgDrDrWsWs wwww Lww',
            [NoWin(),
             Win(limits=0.5)])
        self.scoreTest(
            'RWeWeS1S1B9B9DgDgDrDrWsWwWw LS1',
            [NoWin(6),
             NoWin(limits=0.2)],
            myWind=West, roundWind=North)
        self.scoreTest('RWeWeS2S2B9B9DgDgDrDrWsWsWwWw LS2S2S2', NoWin())
        self.scoreTest('RDbDbDgDgDrDrWsWsWnWnS9C1C1C9 LDrDrDr', NoWin())


class BMJA(Base):

    """specials for chinese classical BMJA"""

    def testMe(self) ->None:
        self.scoreTest(
            'RS1S1S5S6S7S8S9WeWsWwWnS2S3S4 LS3S3',
            [NoWin(),
             Win(limits=1)])


class LastTile(Base):

    """will the best last meld be chosen?"""

    def testMe(self) ->None:
        self.scoreTest(
            'wewewe s1s1s1 b9b9b9 RC1C1C1C2C3 LC1C1C1', [Win(38, 2), Win(34, 2)])
        self.scoreTest(
            'wewewe s1s1s1 b9b9b9 RC1C2C3C3C3 LC3C3C3', [Win(38, 2), Win(34, 2)])
        self.scoreTest(
            'b5b6b7 s1s1s1 RB8C6C7C5B8B8C7C7 mw LC7C7C7', [Win(32, 0), NoWin()])


class CallingHands(Base):

    """diverse calling hands"""

    def testMe(self) ->None:
        self.callingTest('s1s1s1s1 b5b6b7 RB1B8C2C2C6C7C8 Lb5', '')
        self.callingTest('WnWn B1 B2 c4c5c6 b6b6b6 b8b8b8 ye yw', ['b3', ''])
        self.callingTest('WnWn B1 B2 dgdgdg b6b6b6 b8b8b8 ye yw', 'b3')
        self.callingTest('s1s1s1s1 b5b6b7 RB8B8C2C2C6C7C8 Lb5', ['b8c2', ''])
        self.callingTest('s1s1s1s1 b5b6b7 RB7B8C2C2C6C7C8 Lb5', ['b6b9', ''])
        self.callingTest('c3c3c3 RDbDbDbS5S6S7S7S8B2B2 LS8', ['s6s9', ''])
        self.callingTest('RC4C4C5C6C5C7C8 dgdgdg s6s6s6', 'c4c5')
        self.callingTest('RS1S4C5C6C5C7C8 dgdgdg s6s6s6', '')
        self.callingTest(
            'RDbDgDrWsWwWeWnB1B9C1S1S9C9 LWe',
            'dbdgdrwewswwwns1s9b1b9c1c9')


class Recursion(Base):

    """recursion in Hand computing should never happen"""

    def testMe(self) ->None:
        self.scoreTest(
            'c6c6c6C6 fe fs RS8S8C1C2C3C4C5C7C8C9 LC6', [NoWin(16), NoWin(16, 1)])
        self.scoreTest(
            'c6c6c6C6 fe fs RS8S8C1C2C3C4C5C7C8C9 LC7', [NoWin(16), NoWin(16, 1)])


class TstProgram(unittest.TestProgram):

    """we want global access to this program so we can check for verbosity in our tests"""

#    def __init__(self, *args: Any, **kwargs:Mapping[Any, Any]) ->None:
    def __init__(self) ->None:
        global PROGRAM  # pylint: disable=global-statement
        PROGRAM = self
        Debug.callers = '8'
#        unittest.TestProgram.__init__(self, exit=False, *args, **kwargs)
        unittest.TestProgram.__init__(self) #, *args, **kwargs)

if __name__ == '__main__':
    # Debug.hand = True
    TstProgram()



# -*- coding: utf-8 -*-

"""
Copyright (C) 2008-2012 Wolfgang Rohdewald <wolfgang@rohdewald.de>

SPDX-License-Identifier: GPL-2.0-only

"""

# pylint: disable=unused-import

import os
if not os.environ.get('QT_API'):
    os.environ['QT_API'] = 'PYQT6'

# pylint: disable=wrong-import-position

from qtpy import QT6, PYQT6, PYSIDE6

from qtpy.QtCore import QKeyCombination
from qtpy.compat import isalive as qtpy_isalive
from qtpy import QT_VERSION, API_NAME, PYQT_VERSION
from qtpy.QtCore import QAbstractAnimation
from qtpy.QtCore import QAbstractItemModel
from qtpy.QtCore import QAbstractTableModel
from qtpy.QtCore import QByteArray
from qtpy.QtCore import QCoreApplication
from qtpy.QtCore import QCommandLineParser
from qtpy.QtCore import QCommandLineOption
from qtpy.QtCore import QLibraryInfo
from qtpy.QtCore import QStandardPaths
from qtpy.QtCore import QEasingCurve
from qtpy.QtCore import QEvent
from qtpy.QtCore import QEventLoop
from qtpy.QtCore import QItemSelection
from qtpy.QtCore import QMetaObject
from qtpy.QtCore import QMimeData
from qtpy.QtCore import QModelIndex
from qtpy.QtCore import QObject
from qtpy.QtCore import QParallelAnimationGroup
from qtpy.QtCore import QPersistentModelIndex
from qtpy.QtCore import Qt
from qtpy.QtCore import QPoint
from qtpy.QtCore import QPointF
from qtpy.QtCore import QPropertyAnimation
from qtpy.QtCore import QRect
from qtpy.QtCore import QRectF
from qtpy.QtCore import QSize
from qtpy.QtCore import QSizeF
from qtpy.QtCore import QSocketNotifier
from qtpy.QtCore import QTimer
from qtpy.QtCore import QTranslator
from qtpy.QtCore import QLocale
from qtpy.QtCore import Property
from qtpy.QtCore import Signal
from qtpy.QtWidgets import QAbstractItemView
from qtpy.QtWidgets import QAction
from qtpy.QtWidgets import QApplication
from qtpy.QtGui import QBrush
from qtpy.QtGui import QKeyEvent
from qtpy.QtGui import QFocusEvent
from qtpy.QtGui import QHideEvent
from qtpy.QtGui import QShowEvent
from qtpy.QtGui import QMouseEvent
from qtpy.QtGui import QWheelEvent
from qtpy.QtGui import QResizeEvent
from qtpy.QtWidgets import QCheckBox
from qtpy.QtGui import QColor
from qtpy.QtWidgets import QComboBox
from qtpy.QtGui import QCursor
from qtpy.QtWidgets import QDialog
from qtpy.QtWidgets import QDialogButtonBox
from qtpy.QtGui import QDrag
from qtpy.QtGui import QFont
from qtpy.QtGui import QFontMetrics
from qtpy.QtGui import QKeySequence
from qtpy.QtWidgets import QLayout
from qtpy.QtWidgets import QFormLayout
from qtpy.QtWidgets import QFrame
from qtpy.QtWidgets import QGraphicsItem
from qtpy.QtWidgets import QGraphicsItemGroup
from qtpy.QtWidgets import QGraphicsObject
from qtpy.QtWidgets import QGraphicsRectItem
from qtpy.QtWidgets import QGraphicsScene
from qtpy.QtWidgets import QGraphicsSimpleTextItem
from qtpy.QtWidgets import QGraphicsView
from qtpy.QtWidgets import QGraphicsSceneDragDropEvent
from qtpy.QtWidgets import QGridLayout
from qtpy.QtWidgets import QHBoxLayout
from qtpy.QtWidgets import QHeaderView
from qtpy.QtWidgets import QGroupBox
from qtpy.QtGui import QIcon
from qtpy.QtGui import QImageReader
from qtpy.QtCore import QItemSelectionModel
from qtpy.QtWidgets import QLabel
from qtpy.QtWidgets import QLineEdit
from qtpy.QtWidgets import QListWidget
from qtpy.QtWidgets import QListWidgetItem
from qtpy.QtWidgets import QListView
from qtpy.QtWidgets import QMainWindow
from qtpy.QtWidgets import QMenu
from qtpy.QtWidgets import QMessageBox
from qtpy.QtGui import QPainter
from qtpy.QtGui import QPalette
from qtpy.QtGui import QPen
from qtpy.QtGui import QPixmap
from qtpy.QtGui import QPixmapCache
from qtpy.QtWidgets import QProgressBar
from qtpy.QtWidgets import QRadioButton
from qtpy.QtWidgets import QPushButton
from qtpy.QtWidgets import QScrollArea
from qtpy.QtWidgets import QScrollBar
from qtpy.QtWidgets import QSizePolicy
from qtpy.QtWidgets import QSlider
from qtpy.QtWidgets import QSpacerItem
from qtpy.QtWidgets import QSpinBox
from qtpy.QtWidgets import QSplitter
from qtpy.QtWidgets import QStackedWidget
from qtpy.QtWidgets import QStatusBar
from qtpy.QtCore import QStringListModel
from qtpy.QtWidgets import QStyle
from qtpy.QtWidgets import QStyledItemDelegate
from qtpy.QtWidgets import QStyleOption
from qtpy.QtWidgets import QStyleOptionGraphicsItem
from qtpy.QtWidgets import QStyleOptionViewItem
from qtpy.QtWidgets import QTableView
from qtpy.QtWidgets import QTableWidget
from qtpy.QtWidgets import QTableWidgetItem
from qtpy.QtWidgets import QTabWidget
from qtpy.QtWidgets import QTextBrowser
from qtpy.QtGui import QTextDocument
from qtpy.QtWidgets import QTextEdit
from qtpy.QtWidgets import QToolBar
from qtpy.QtWidgets import QToolButton
from qtpy.QtGui import QTransform
from qtpy.QtWidgets import QTreeView
from qtpy.QtWidgets import QVBoxLayout
from qtpy.QtWidgets import QWidget
from qtpy.QtGui import QValidator
from qtpy.QtGui import QGuiApplication
from qtpy.QtSvgWidgets import QGraphicsSvgItem  # type: ignore
from qtpy.QtSvg import QSvgRenderer


HAVE_SIP = True
if PYQT6:
    from PyQt6 import sip  # type:ignore[no-redef]
    def sip_cast(obj, _type):
        """hide not so nice things in qt.py"""
        return sip.cast(obj, _type)
else:
    HAVE_SIP = False
    def sip_cast(obj, _type):
        """hide not so nice things in qt.py"""
        raise NotImplementedError

def modeltest_is_supported() ->bool:
    """Is the QT binding supported."""
    if not HAVE_SIP:
        return False
    try:
        _ = sip_cast(QSize(), QSize)
        return True
    except (TypeError, NotImplementedError):
        return False

SIP_VERSION_STR = 'no sip'
if HAVE_SIP:
    SIP_VERSION_STR = sip.SIP_VERSION_STR  # type:ignore[name-defined]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Copyright (C) 2016 Wolfgang Rohdewald <wolfgang@rohdewald.de>

SPDX-License-Identifier: GPL-2.0-only

"""

import signal
import os
import sys
import subprocess
import random
import shutil
import time
import gc

import argparse
from locale import getpreferredencoding

from typing import List, Set, Optional, Any, Generator, Iterable, Union, TYPE_CHECKING, Tuple, cast

from common import Debug, ReprMixin, cacheDir
from util import removeIfExists, gitHead, checkMemory, popenReadlines
from kajcsv import Csv, CsvRow, CsvWriter

if TYPE_CHECKING:
    from io import FileIO


signal.signal(signal.SIGINT, signal.SIG_DFL)


OPTIONS: Any

class Clone:

    """make a temp directory for commitId"""

    def __init__(self, commitId:str) ->None:
        self.commitId = commitId
        if commitId != 'current':
            tmpdir = os.path.expanduser(os.path.join(cacheDir(), commitId))
            if not os.path.exists(tmpdir):
                with subprocess.Popen(
                    f'git clone --shared --no-checkout -q .. {tmpdir}'.split()) as _:
                    _.wait()
                with subprocess.Popen(
                    f'git checkout -q {commitId}'.split(), cwd=tmpdir) as _:
                    _.wait()

    def sourceDirectory(self) ->str:
        """the source directory for this git commit"""
        if self.commitId == 'current':
            tmpdir = os.path.abspath('..')
            result = os.path.join(tmpdir, 'src')
        else:
            result = os.path.join(cacheDir(), self.commitId, 'src')
        assert os.path.exists(result), f'{result} does not exist'
        return result

    @classmethod
    def removeObsolete(cls, knownCommits:Set) ->None:
        """remove all clones for obsolete commits"""
        for commitDir in os.listdir(cacheDir()):
            if not any(x.startswith(commitDir) for x in knownCommits):
                removeDir = os.path.join(cacheDir(), commitDir)
                shutil.rmtree(removeDir)


class TooManyClients(UserWarning):

    """we would surpass options.clients"""


class TooManyServers(UserWarning):

    """we would surpass options.servers"""


class Server(ReprMixin):

    """represents a kajongg server instance. Called when we want to start a job."""
    servers:List['Server'] = []
    count = 0

    def __new__(cls, job:'Job') ->'Server':
        """can we reuse an existing server?"""
        running = Server.allRunningJobs()
        if len(running) >= OPTIONS.clients:
            raise TooManyClients
        maxClientsPerServer = OPTIONS.clients / OPTIONS.servers
        matchingServers = [
            x for x in cls.servers
            if x.commitId == job.commitId
            and x.pythonVersion == job.pythonVersion
            and len(x.jobs) < maxClientsPerServer]
        if matchingServers:
            result = sorted(matchingServers, key=lambda x: len(x.jobs))[0]
        else:
            if len(cls.servers) >= OPTIONS.servers:
                # maybe we can kill a server without jobs?
                for server in cls.servers:
                    if not server.jobs:
                        server.stop()
                        break  # we only need to stop one server
                else:
                    # no server without jobs found
                    raise TooManyServers
            result = object.__new__(cls)
            cls.servers.append(result)
        return result

    def __init__(self, job:'Job') ->None:
        if not hasattr(self, 'jobs'):
            self.jobs = []
            self.process:Optional[subprocess.Popen] = None
            self.socketName:Optional[str] = None
            self.portNumber:Optional[int] = None
            self.commitId:str = job.commitId
            self.pythonVersion:str = job.pythonVersion
            self.clone = Clone(job.commitId)
            self.start(job)
        else:
            self.jobs.append(job)

    @classmethod
    def allRunningJobs(cls) ->List['Job']:
        """a list of all jobs on all servers"""
        return sum((x.jobs for x in cls.servers), [])

    def start(self, job:'Job') ->None:
        """start this server"""
        job.server = self
        assert self.process is None, 'Server.start already has a process'
        self.jobs.append(job)
        assert self.commitId == job.commitId, f'Server.commitId {self.commitId} != Job.commitId {job.commitId}'
        cmd = [os.path.join(
            job.srcDir(),
            'kajonggserver.py')]
        cmd.insert(0, f'python{self.pythonVersion}')
        if OPTIONS.usePort:
            self.portNumber = random.randrange(1025, 65000)
            cmd.append(f'--port={self.portNumber}')
        else:
            Server.count += 1
            self.socketName = os.path.expanduser(
                os.path.join('~', '.kajongg',
                             f'sock{self.commitId}.py{self.pythonVersion}.{Server.count}'))
            cmd.append(f'--socket={self.socketName}')
        if OPTIONS.debug:
            cmd.append(f"--debug={','.join(OPTIONS.debug)}")
        if OPTIONS.log:
            self.process = subprocess.Popen(
                cmd, cwd=job.srcDir(),
                stdout=job.logFile, stderr=job.logFile)
        else:
            # reuse this server (otherwise it stops by itself)
            cmd.append('--continue')
            self.process = subprocess.Popen(cmd, cwd=job.srcDir())  # pylint:disable=consider-using-with
        print(f'{self} started')

    def stop(self, job:Optional['Job']=None) ->None:
        """maybe stop the server"""
        if self not in self.servers:
            # already stopped
            return
        if job:
            self.jobs.remove(job)
        if not self.jobs:
            self.servers.remove(self)
            if self.process:
                try:
                    self.process.terminate()
                    self.process.wait()
                except OSError:
                    pass
                print(f'{self} killed')
            if self.socketName:
                removeIfExists(self.socketName)

    @classmethod
    def stopAll(cls) ->None:
        """stop all servers even if clients are still there"""
        for server in cls.servers:
            for job in server.jobs[:]:
                server.stop(job)
            assert not server.jobs, f'stopAll expects no server jobs but found {server.jobs}'
            server.stop()

    def __str__(self) ->str:
        return 'Server {} Python{}{} {}'.format( # pylint:disable=consider-using-f-string
            self.commitId,
            self.pythonVersion,
            ' pid={}'.format(self.process.pid) if Debug.process and self.process else '', # pylint:disable=consider-using-f-string
            ' port={}'.format(self.portNumber) if self.portNumber else ' socket={}'.format(self.socketName)) # pylint:disable=consider-using-f-string


class Job(ReprMixin):

    """a simple container"""

    def __init__(self, pythonVersion:str, ruleset:str, aiVariant:str, commitId:str, game:str) ->None:
        self.pythonVersion = pythonVersion
        self.ruleset = ruleset
        self.aiVariant = aiVariant
        self.commitId = commitId
        self.game = game
        self.__logFile:'FileIO'
        self.logFileName:str
        self.process:Optional[subprocess.Popen] = None
        self.server:Server
        self.started = False

    def srcDir(self) ->str:
        """the path of the directory where the particular test is running"""
        assert self.server, f'Job {self} has no server'
        assert self.server.clone, f'Job {self} has no server.clone'
        return self.server.clone.sourceDirectory()

    def __startProcess(self, cmd:List[str]) ->None:
        """call Popen"""
        if OPTIONS.log:
            self.process = subprocess.Popen(
                cmd, cwd=self.srcDir(),
                stdout=self.logFile, stderr=self.logFile)
        else:
            self.process = subprocess.Popen(cmd, cwd=self.srcDir())  # pylint:disable=consider-using-with
        print(f'       {self} started')

    def start(self) ->None:
        """start this job"""
        self.server = Server(self)
        # never login to the same server twice at the
        # same time with the same player name
        player = self.server.jobs.index(self) + 1
        cmd = [os.path.join(self.srcDir(), 'kajongg.py'),
               f'--game={self.game}',
               f'--player=Tüster {player}',
               f'--ruleset={self.ruleset}']
        if self.server.socketName:
            cmd.append(f'--socket={self.server.socketName}')
        if self.server.portNumber:
            cmd.append(f'--port={self.server.portNumber}')
        cmd.insert(0, f'python{self.pythonVersion}')
        if self.aiVariant != 'DefaultAI':
            cmd.append(f'--ai={self.aiVariant}')
        if OPTIONS.csv:
            cmd.append(f'--csv={OPTIONS.csv}')
        if OPTIONS.gui:
            cmd.append('--demo')
        else:
            cmd.append('--nogui')
        if OPTIONS.playopen:
            cmd.append('--playopen')
        if OPTIONS.debug:
            cmd.append(f"--debug={','.join(OPTIONS.debug)}")
        self.__startProcess(cmd)
        self.started = True

    def check(self, silent:bool=False) ->None:
        """if done, cleanup"""
        if not self.started or not self.process:
            return
        result:Union[int, str, None] = self.process.poll()
        if result is not None:
            assert isinstance(result, int)
            self.process = None
            if not silent:
                if result < 0:
                    result = signal.Signals(-result).name
                if result:
                    result = f' with return {result}'
                else:
                    result = ''
                print(f'       {self} done{result}')
            self.server.jobs.remove(self)

    @property
    def logFile(self) ->'FileIO':
        """open if needed"""
        if not hasattr(self, '__logFile'):
            logDir = os.path.expanduser(
                os.path.join('~', '.kajongg', 'log', str(self.game),
                             self.ruleset, self.aiVariant, str(self.pythonVersion)))
            if not os.path.exists(logDir):
                os.makedirs(logDir)
            logFileName = self.commitId
            self.logFileName = os.path.join(logDir, logFileName)
            self.__logFile = open(self.logFileName, 'ab', buffering=0)  # pylint:disable=consider-using-with
        return self.__logFile

    def shortRulesetName(self) ->str:
        """strip leading chars if they are identical for all rulesets"""
        names = OPTIONS.knownRulesets
        for prefix in range(100):
            if sum(x.startswith(self.ruleset[:prefix]) for x in names) == 1:
                return self.ruleset[prefix - 1:]
        return self.ruleset

    def __str__(self) ->str:
        pid = f'pid={self.process.pid}' if self.process and Debug.process else ''
        game = f'game={self.game}'
        ruleset = self.shortRulesetName()
        aiName = f'AI={self.aiVariant}' if self.aiVariant != 'DefaultAI' else ''
        return ' '.join([
            self.commitId, f'Python{self.pythonVersion}', pid, game, ruleset, aiName]).replace('  ', ' ')




def cleanup_data(csv:'CSV') ->None:
    """remove all data referencing obsolete commits"""
    logDir = os.path.expanduser(os.path.join('~', '.kajongg', 'log'))
    knownCommits = csv.commits()
    for dirName, _, fileNames in os.walk(logDir):
        for fileName in fileNames:
            if fileName not in knownCommits and fileName != 'current':
                os.remove(os.path.join(dirName, fileName))
        try:
            os.removedirs(dirName)
        except OSError:
            pass  # not yet empty
    Clone.removeObsolete(knownCommits)

def pairs(data:List[CsvRow]) ->Generator[Tuple[CsvRow, CsvRow], None, None]:
    """return all consecutive pairs"""
    prev = None
    for _ in data:
        if prev:
            yield prev, _
        prev = _


class CSV(ReprMixin):
    """represent kajongg.csv"""

    knownCommits : Set[str] = set()

    def __init__(self) ->None:
        self.findKnownCommits()
        self.rows:List[CsvRow] = []
        if os.path.exists(OPTIONS.csv):
            self.rows = list(sorted({CsvRow(x) for x in Csv.reader(OPTIONS.csv)}))
        self.removeInvalidCommits()

    def neutralize(self) ->None:
        """remove things we do not want to compare"""
        for row in self.rows:
            row.neutralize()

    def commits(self) ->Set[str]:
        """return set of all our commit ids"""
        # TODO: sorted by date
        return {x.commit for x in self.rows}

    def games(self) ->List[str]:
        """return a sorted unique list of all games"""
        return sorted({x.game for x in self.rows})

    @classmethod
    def findKnownCommits(cls) ->None:
        """find known commits"""
        if not cls.knownCommits:
            cls.knownCommits = set()
            for branch in subprocess.check_output(b'git branch'.split()).decode().split('\n'):
                if 'detached' not in branch and 'no branch' not in branch:
                    cls.knownCommits |= set(subprocess.check_output(
                        f'git log --max-count=800 --pretty=%H {branch[2:]} --abbrev=10'.split()).decode().split('\n'))

    @classmethod
    def onlyExistingCommits(cls, commits:Iterable[str]) ->Set[str]:
        """return a set with only  existing commits"""
        result = set()
        for commit in commits:
            if any(x.startswith(commit) for x in cls.knownCommits):
                result.add(commit)
        return result

    def removeInvalidCommits(self) ->None:
        """remove rows with invalid git commit ids"""
        csvCommits = {x.commit for x in self.rows}
        csvCommits = {
            x for x in csvCommits if set(
                x) <= set(
                    '0123456789abcdef') and len(
                        x) >= 7}
        nonExisting = csvCommits - self.onlyExistingCommits(set(x.commit for x in self.rows))
        if nonExisting:
            print(
                f"removing rows from kajongg.csv for commits {','.join(nonExisting)}")
            self.rows = [x for x in self.rows if x.commit not in nonExisting]
            self.write()

    def write(self) ->None:
        """write new csv file"""
        writer = CsvWriter(OPTIONS.csv)
        for row in self.rows:
            writer.writerow(row)
        del writer

    def evaluate(self) ->None:
        """evaluate the data. Show differences as helpful as possible"""
        found_difference = False
        for ruleset, aiVariant, game in {(x.ruleset, x.aiVariant, x.game) for x in self.rows}:
            rows = list(reversed(sorted(
                x for x in self.rows
                if ruleset == x.ruleset and aiVariant == x.aiVariant and game == x.game)))
            for fixedField in (CsvRow.Fields.PY_VERSION, CsvRow.Fields.COMMIT):
                for fixedValue in set(x[fixedField] for x in rows):
                    checkRows = [x for x in rows if x[fixedField] == fixedValue]
                    for warned in self.compareRows(checkRows):
                        found_difference = True
                        rows.remove(warned)
            found_difference |= len(self.compareRows(rows)) > 0
            self.compareRows(rows)
        if not found_difference:
            print(f'found no differences in {OPTIONS.csv}')

    @staticmethod
    def compareRows(rows:List[CsvRow]) ->List[CsvRow]:
        """in absence of differences, there should be only one row.
        return a list of rows which appeared in warnings"""
        if not rows:
            return []
        msgHeader = None
        result:List[CsvRow] = []
        ruleset = rows[0].ruleset
        aiVariant = rows[0].aiVariant
        game = rows[0].game
        differences:List[Tuple[CsvRow, CsvRow, Tuple[str, str]]] = []
        for pair in pairs(rows):
            causes = pair[1].differs_for(pair[0])
            if causes:
                triple = cast(Tuple[CsvRow, CsvRow, Tuple[str, str]], tuple([pair[0], pair[1], causes]))
                differences.append(triple)
        for difference in sorted(differences, key=lambda x: len(x[2])):
            if not set(difference[:2]) & set(result):
                if msgHeader is None:
                    msgHeader = f'looking at game={game} ruleset={ruleset} AI={aiVariant}'
                    print(msgHeader)
                print('   {} {}'.format(*difference[2])) # pylint:disable=consider-using-f-string
                result.extend(difference[:2])
        return result


def startingDir() ->str:
    """the path of the directory where kajonggtest has been started in"""
    return os.path.dirname(sys.argv[0])


def getJobs(jobs:List[Job]) ->List[Job]:
    """fill the queue"""
    try:
        while len(jobs) < OPTIONS.clients:
            jobs.append(next(OPTIONS.jobs))
    except StopIteration:
        pass
    return jobs


def doJobs() ->None:
    """now execute all jobs"""
    # pylint: disable=too-many-branches

    if not OPTIONS.git and OPTIONS.csv:
        if gitHead() in ('current', None):
            print(
                f"Disabling CSV output: {'You have uncommitted changes' if gitHead() == 'current' else 'No git'}")
            print()
            OPTIONS.csv = None

    try:
        jobs:List[Job] = []
        while getJobs(jobs):
            for checkJob in Server.allRunningJobs()[:]:
                checkJob.check()
            try:
                jobs[0].start()
                jobs = jobs[1:]
            except TooManyServers:
                time.sleep(3)
            except TooManyClients:
                time.sleep(3)
    except KeyboardInterrupt:
        Server.stopAll()
    except BaseException as exc:
        print(exc)
        raise exc
    finally:
        while True:
            running = Server.allRunningJobs()
            if not running:
                break
            for job in running:
                if not job.started:
                    if job.server:
                        job.server.jobs.remove(job)
                else:
                    job.check()
                    if job.process:
                        print(f'Waiting for   {job}')
                        job.process.wait()
            time.sleep(1)


def parse_options() ->argparse.Namespace:
    """parse options"""
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '--gui', dest='gui', action='store_true',
        default=False, help='show graphical user interface')
    parser.add_argument(
        '--ruleset', dest='rulesets', default='ALL',
        help='play like a robot using RULESET: comma separated list. If missing, test all rulesets',
        metavar='RULESET')
    parser.add_argument(
        '--ai', dest='aiVariants',
        default=None, help='use AI variants: comma separated list',
        metavar='AI')
    parser.add_argument(
        '--python', dest='pyVersions',
        default=None, help='use python versions: comma separated list',
        metavar='PY_VERSION')
    parser.add_argument(
        '--log', dest='log', action='store_true',
        default=False, help='write detailled debug info to ~/.kajongg/log/game/ruleset/commit.'
                            ' This starts a separate server process per job, it sets --servers to --clients.')
    parser.add_argument(
        '--game', dest='game',
        help='start first game with GAMEID, increment for following games.' +
        ' Without this, random values are used.',
        metavar='GAMEID', type=int, default=0)
    parser.add_argument(
        '--count', dest='count',
        help='play COUNT games. Default is unlimited',
        metavar='COUNT', type=int, default=999999999)
    parser.add_argument(
        '--playopen', dest='playopen', action='store_true',
        help='all robots play with visible concealed tiles', default=False)
    parser.add_argument(
        '--clients', dest='clients',
        help='start a maximum of CLIENTS kajongg instances. Default is 2',
        metavar='CLIENTS', type=int, default=2)
    parser.add_argument(
        '--servers', dest='servers',
        help='start a maximum of SERVERS kajonggserver instances. Default is 1',
        metavar='SERVERS', type=int, default=1)
    parser.add_argument(
        '--git', dest='git',
        help='check all commits: either a comma separated list or a range from..until')
    parser.add_argument(
        '--debug', dest='debug',
        help=Debug.help())
    return parser.parse_args()


def improve_options() ->None:
    """add sensible defaults"""
    # pylint: disable=too-many-branches,too-many-statements
    OPTIONS.servers = max(OPTIONS.servers, 1)

    cmdPath = os.path.join(startingDir(), 'kajongg.py')
    cmd = ['python3', cmdPath, '--rulesets']
    OPTIONS.knownRulesets = list(popenReadlines(cmd))
    if OPTIONS.rulesets == 'ALL':
        OPTIONS.rulesets = OPTIONS.knownRulesets
    else:
        wantedRulesets = OPTIONS.rulesets.split(',')
        usingRulesets = []
        wrong = False
        for ruleset in wantedRulesets:
            matches = [x for x in OPTIONS.knownRulesets if ruleset in x]
            if not matches:
                print('ruleset', ruleset, 'is not known', end=' ')
                wrong = True
            elif len(matches) > 1:
                exactMatch = [x for x in OPTIONS.knownRulesets if ruleset == x]
                if len(exactMatch) == 1:
                    usingRulesets.append(exactMatch[0])
                else:
                    print('ruleset', ruleset, 'is ambiguous:', matches)
                    wrong = True
            else:
                usingRulesets.append(matches[0])
        if wrong:
            sys.exit(1)
        OPTIONS.rulesets = usingRulesets
    if OPTIONS.git is not None:
        if '..' in OPTIONS.git:
            if '^' not in OPTIONS.git:
                OPTIONS.git = OPTIONS.git.replace('..', '^..')
            commits = subprocess.check_output(
                f'git log --grep _SILENT --invert-grep --pretty=%h {OPTIONS.git} --abbrev=10'.split()).decode()
            _ = list(x.strip() for x in commits.split('\n') if x.strip())
            OPTIONS.git = list(reversed(_))
        else:
            OPTIONS.git = CSV.onlyExistingCommits(OPTIONS.git.split(','))
            if not OPTIONS.git:
                sys.exit(1)
    if OPTIONS.debug is None:
        OPTIONS.debug = []
    else:
        OPTIONS.debug = [OPTIONS.debug]
    if OPTIONS.log:
        OPTIONS.servers = OPTIONS.clients
        OPTIONS.debug.extend(
            'neutral,dangerousGame,explain,originalCall,robbingKong,robotAI,scores,traffic,hand'.split(','))
    if gitHead() not in ('current', None) and not OPTIONS.log:
        OPTIONS.debug.append('git')
    if not OPTIONS.aiVariants:
        OPTIONS.aiVariants = 'DefaultAI'
    if OPTIONS.pyVersions:
        OPTIONS.pyVersions = OPTIONS.pyVersions.split(',')
    else:
        OPTIONS.pyVersions = ['3']
    OPTIONS.allAis = OPTIONS.aiVariants.split(',')
    if OPTIONS.count:
        print('rulesets:', ', '.join(OPTIONS.rulesets))
        _ = ' '.join(OPTIONS.allAis)
        if _ != 'DefaultAI':
            print('AIs:', _)
    if OPTIONS.git:
        print('commits:', ' '.join(OPTIONS.git))
        # since we order jobs by game, commit we want one permanent server per
        # commit
    OPTIONS.jobs = allJobs()
    OPTIONS.games = allGames()
    OPTIONS.jobCount = 0
    OPTIONS.usePort = sys.platform == 'win32'


def allGames() ->Generator[str, None, None]:
    """a generator returning game ids"""
    while True:
        if OPTIONS.game:
            result = OPTIONS.game
            OPTIONS.game += 1
        else:
            result = int(random.random() * 10 ** 9)
        yield result


def allJobs() ->Generator[Job, None, None]:
    """a generator returning Job instances"""
    # pylint: disable=too-many-nested-blocks
    for game in OPTIONS.games:
        for commitId in OPTIONS.git or ['current']:
            for ruleset in OPTIONS.rulesets:
                for aiVariant in OPTIONS.allAis:
                    for pyVersion in OPTIONS.pyVersions:
                        OPTIONS.jobCount += 1
                        if OPTIONS.jobCount > OPTIONS.count:
                            return
                        yield Job(pyVersion, ruleset, aiVariant, commitId, game)

def main() ->None:
    """parse options, play, evaluate results"""
    global OPTIONS  # pylint: disable=global-statement

    locale_encoding = getpreferredencoding()
    if locale_encoding.lower() != 'utf-8':
        print(f'we need default encoding utf-8 but have {locale_encoding}')
        sys.exit(2)

    # we want only english in the logs because i18n and friends
    # behave differently in kde and kde
    os.environ['LANG'] = 'en_US.UTF-8'
    OPTIONS = parse_options()
    OPTIONS.csv = os.path.expanduser(
        os.path.join('~', '.kajongg', 'kajongg.csv'))
    if not os.path.exists(os.path.dirname(OPTIONS.csv)):
        os.makedirs(os.path.dirname(OPTIONS.csv))

    csv = CSV()

    improve_options()

    csv.evaluate()
    cleanup_data(csv)

    errorMessage = Debug.setOptions(','.join(OPTIONS.debug))
    if errorMessage:
        print(errorMessage)
        sys.exit(2)

    print()

    if OPTIONS.count:
        doJobs()
        if OPTIONS.csv:
            CSV().evaluate()

def cleanup(sig: Any, unusedFrame: Any) ->None:
    """at program end"""
    Server.stopAll()
    sys.exit(sig)

# is one server for two clients.
if __name__ == '__main__':
    signal.signal(signal.SIGABRT, cleanup)
    signal.signal(signal.SIGTERM, cleanup)
    signal.signal(signal.SIGINT, cleanup)
    if sys.platform != 'win32':
        signal.signal(signal.SIGHUP, cleanup)
        signal.signal(signal.SIGQUIT, cleanup)

    main()
    gc.collect()
    checkMemory()



# -*- coding: utf-8 -*-

"""
Copyright (C) 2008-2016 Wolfgang Rohdewald <wolfgang@rohdewald.de>

SPDX-License-Identifier: GPL-2.0-only

"""

from collections import defaultdict
import datetime
import sys
import os
import shutil
import logging
import logging.handlers
import socket
import string
from signal import signal, SIGABRT, SIGINT, SIGTERM
from typing import Optional, Any, Union, List
from typing import TYPE_CHECKING, Iterable, Generator, Literal, cast

from qt import QStandardPaths, QObject, modeltest_is_supported, qtpy_isalive

if TYPE_CHECKING:
    from tile import Tile
    from mainwindow import MainWindow
    from twisted.internet.interfaces import IReactorCore
    from config import SetupPreferences
    from qt import QGraphicsItem
    from scene import GameScene

# pylint: disable=invalid-name

if sys.platform == 'win32':
    # This is only needed for manual execution, and
    # we expect python to be the python3 interpreter.
    # The windows installer will use kajongg.exe and kajonggserver.exe
    interpreterName = 'python'
else:
    interpreterName = 'python3'

# TODO: enumeration
LIGHTSOURCES = cast(Union[Literal['NE'], Literal['NW'], Literal['SW'], Literal['SE']], ['NE', 'NW', 'SW', 'SE'])


def isAlive(qobj: Union[QObject, 'QGraphicsItem', None]) ->bool:
    """check if the underlying C++ object still exists"""
    if qobj is None:
        return False
    result = qtpy_isalive(qobj)
    if not result and Debug.isalive:
        print('NOT alive:', repr(qobj))
    return result

def serverAppdataDir() ->str:
    """
    The per user directory with kajongg application information like the database.

    @return: The directory path.
    @rtype: C{str}.
    """
    serverDir = os.path.expanduser('~/.kajonggserver/')
    # the server might or might not have KDE installed, so to be on
    # the safe side we use our own .kajonggserver directory
    # the following code moves an existing kajonggserver.db to .kajonggserver
    # but only if .kajonggserver does not yet exist
    kdehome = os.environ.get('KDEHOME', '~/.kde')
    oldPath = os.path.expanduser(
        f"{kdehome}/share/apps/kajongg/kajonggserver.db")
    if not os.path.exists(oldPath):
        oldPath = os.path.expanduser(
            '~/.kde4/share/apps/kajongg/kajonggserver.db')
    if os.path.exists(oldPath) and not os.path.exists(serverDir):
        # upgrading an old kajonggserver installation
        os.makedirs(serverDir)
        shutil.move(oldPath, serverDir)
    if not os.path.exists(serverDir):
        try:
            os.makedirs(serverDir)
        except OSError:
            pass
    return serverDir


def clientAppdataDir() ->str:
    """
    The per user directory with kajongg application information like the database.

    @return: The directory path.
    @rtype: C{str}.
    """
    serverDir = os.path.expanduser('~/.kajonggserver/')
    if not os.path.exists(serverDir):
        # the client wants to place the socket in serverDir
        os.makedirs(serverDir)
    result = QStandardPaths.writableLocation(QStandardPaths.StandardLocation.AppDataLocation)
    # this may end with kajongg.py or .pyw or whatever, so fix that:
    if not os.path.isdir(result):
        result = os.path.dirname(result)
    if not result.endswith('kajongg'):
        # when called first, QApplication.applicationName is not yet set
        result = result + '/kajongg'
    if not os.path.exists(result):
        os.makedirs(result)
    return result


def appdataDir() ->str:
    """
    The per user directory with kajongg application information like the database.

    @return: The directory path.
    @rtype: C{str}.
    """
    return serverAppdataDir() if Internal.isServer else clientAppdataDir()


def cacheDir() ->str:
    """the cache directory for this user"""
    cachedir = os.environ.get('XDG_CACHE_HOME', '~/.cache')
    result = os.path.expanduser(cachedir + '/kajongg')
    if not os.path.exists(result):
        os.makedirs(result)
    return result


def socketName() ->str:
    """client and server process use this socket to talk to each other"""
    serverDir = os.path.expanduser('~/.kajonggserver')
    if not os.path.exists(serverDir):
        appdataDir()
                   # allocate the directory and possibly move old databases
                   # there
    if Options.socket:
        return Options.socket
    return os.path.normpath(f'{serverDir}/socket{Internal.defaultPort}')


def handleSignals(handler: Any) ->None:

    """set up signal handling"""

    signal(SIGABRT, handler)
    signal(SIGINT, handler)
    signal(SIGTERM, handler)
    if sys.platform != 'win32':
        from signal import SIGHUP, SIGQUIT
        signal(SIGHUP, handler)
        signal(SIGQUIT, handler)


class Debug:

    """holds flags for debugging output. At a later time we might
    want to add command line parameters for initialisation, and
    look at kdebugdialog"""
    connections = False
    traffic = False
    process = False
    time = False
    timestamp:Optional[datetime.datetime] = None
    sql = False
    animation = ''  # 'yeysywynG87gfefsfwfn' for tiles and G#g for groups where # is the uid
    animationSpeed = False
    robotAI = False
    dangerousGame = False
    originalCall = False
    modelTest = False
    focusable = ''
    robbingKong = False
    mahJongg = False
    sound = False
    chat = False
    argString = None
    scores = False
    hand = False
    explain = False
    random = False
    deferredBlock = False
    stack = False
    events = ''
    table = False
    gc = False
    delayChow = False
    locate = False
    neutral = False  # only neutral comparable debug output
    callers = '0'
    git = False
    ruleCache = False
    quit = False
    preferences = False
    graphics = False
    scoring = False
    wallSize = '0'
    i18n = False
    isalive = False
    uitiles = False

    def __init__(self) ->None:
        raise TypeError('Debug is not meant to be instantiated')

    @staticmethod
    def help() ->'str':
        """a string for help texts about debug options"""
        def optYielder(options: List[str]) -> Generator[str, None, None]:
            """yields options with markers for line separation"""
            for idx, opt in enumerate(options):
                yield opt
                if idx < len(options) - 1 and idx % 5 == 4:
                    yield 'SEPARATOR'
        options = [x for x in Debug.__dict__ if not x.startswith('_')]
        boolOptions = sorted(x for x in options
                             if isinstance(Debug.__dict__[x], bool))
        stringOptions = sorted(x for x in options
                               if isinstance(Debug.__dict__[x], str))
        stringExample = f'{stringOptions[0]}:s3s4'
        allOptions = sorted(boolOptions + stringOptions)
        opt = '\n'.join(
            ', '.join(optYielder(allOptions)).split(' SEPARATOR, '))
        # TODO: i18n for this string. First move i18n out of kde so we can import it here
        return f"""set debug options. Pass a comma separated list of options.
Options are: {opt}.
Options {', '.join(stringOptions)} take a string argument like {stringExample}.
--debug=events can get suboptions like in --debug=events:Mouse:Hide
     showing all event messages with 'Mouse' or 'Hide' in them"""

    @staticmethod
    def setOptions(args: 'str') ->'str':
        """args comes from the command line. Put this in the Debug class.
        If something goes wrong, return an error message."""
        if not args:
            return ''
        Debug.argString = args
        for arg in args.split(','):
            parts = arg.split(':')
            option = parts[0]
            value: Union[bool, str]
            if len(parts) == 1:
                value = True
            else:
                value = ':'.join(parts[1:])
            if option not in Debug.__dict__:
                return f'--debug: unknown option {option}'
            if not isinstance(Debug.__dict__[option], type(value)):
                return (f"--debug: wrong type for option {option}: "
                        f"given {value}/{type(value)}, should be {type(Debug.__dict__[option])}")
            if option != 'scores' or not Internal.isServer:
                type.__setattr__(Debug, option, value)
        if Debug.time:
            Debug.timestamp = datetime.datetime.now()
        if Debug.modelTest and not modeltest_is_supported():
            print('--debug=modelTest is not yet supported for pyside, use pyqt')
            sys.exit(2)
        return ''

    @staticmethod
    def str() ->str:
        """__str__ does not work with class objects"""
        result = []
        for option in Debug.__dict__:
            if not option.startswith('_'):
                result.append(f'{option}={getattr(Debug, option)}')
        return ' '.join(result)


class FixedClass(type):

    """Metaclass: after the class variable fixed is set to True,
    all class variables become immutable"""
    def __setattr__(cls, key: str, value: object) ->None:
        if cls.fixed:  # type: ignore
            raise SystemExit(f'{cls.__name__}.{key} may not be changed')
        type.__setattr__(cls, key, value)


def num_encode(number: int, length: int=4) ->str:
    """make a short unique ascii string out of number, truncate to length"""
    alphabet = string.ascii_uppercase + string.ascii_lowercase
    base = len(alphabet)
    result : List[str] = []
    while number and len(result) < length:
        number, remainder = divmod(number, base)
        result.append(alphabet[remainder])
    return ''.join(reversed(result))


def id4(obj: object) ->str:
    """object id for debug messages"""
    if obj is None:
        return 'NONE'
    try:
        if hasattr(obj, 'uid'):
            return obj.uid
    except Exception:  # pylint: disable=broad-except
        pass
    if Debug.neutral:
        return '.'
    return num_encode(id(obj))


class ReprMixin:

    """
    A mixin defining a default for __repr__,
    using __str__. If __str__ is not defined, this runs
    into recursion. But I see no easy way without too much
    runtime overhead to check for this beforehand.
    """

    def __repr__(self) ->str:
        clsName = self.__class__.__name__
        content = str(self)
        if content.startswith(clsName):
            return content
        return f'{clsName}_{id4(self)}({content})'


class Options(metaclass=FixedClass):

    """they are never saved in a config file. Some of them
    can be defined on the command line."""
    demo = False
    showRulesets = False
    rulesetName = None  # will only be set by command line --ruleset
    ruleset = None       # from rulesetName
    host = None
    player = None
    dbPath = None
    socket : Optional['str'] = None
    port = None
    playOpen = False
    gui = False
    AI = 'DefaultAI'
    csv = None
    continueServer = False
    fixed = False

    def __init__(self) ->None:
        raise TypeError('Options is not meant to be instantiated')

    @staticmethod
    def str() -> str:
        """__str__ does not work with class objects"""
        result = []
        for option in Options.__dict__:
            if not option.startswith('_'):
                value = getattr(Options, option)
                if isinstance(value, (bool, int, str)):
                    result.append(f'{option}={value}')
        return ' '.join(result)

class SingleshotOptions:

    """Options which are cleared after having been used once"""
    table = False
    join = False
    game = None


class __Internal:

    """
    Global things.

    @cvar Preferences: The L{SetupPreferences}.
    @type Preferences: L{SetupPreferences}
    @cvar version: The version of Kajongg.
    @type version: C{str}
    @cvar logPrefix: C for client and S for server.
    @type logPrefix: C{str}
    @cvar isServer: True if this is the server process.
    @type isServer: C{bool}
    @cvar reactor: The twisted reactor instance.
    @type reactor: L{twisted.internet.reactor}
    @cvar app: The Qt or KDE app instance
    @type app: L{KApplication}
    @cvar db: The sqlite3 data base
    @type db: L{DBHandle}
    @cvar scene: The game scene.
    @type scene: L{GameScene}: L{PlayingScene} or L{ScoringScene}
    """
    Preferences:Optional['SetupPreferences'] = None
    defaultPort = 8303
    logPrefix = 'C'
    isServer = False
    reactor:'IReactorCore'
    app : Any = None
    db : Any = None
    scene:Optional['GameScene'] = None
    mainWindow : Optional['MainWindow'] = None
    game = None
    autoPlay = False
    logger : Any = None
    kajonggrc : Any = None

    def __init__(self) ->None:
        """init the loggers"""
        handler: Any
        logName = f"{os.path.basename(sys.argv[0]).replace('.py', '').replace('.exe', '')}.log"
        self.logger = logging.getLogger(logName)
        if sys.platform == 'win32':
            haveDevLog = False
        else:
            try:
                handler = logging.handlers.SysLogHandler('/dev/log')
                haveDevLog = True
            except (AttributeError, socket.error):
                haveDevLog = False
        if not haveDevLog:
            logName = os.path.join(appdataDir(), logName)
            handler = logging.handlers.RotatingFileHandler(
                logName, maxBytes=100000000, backupCount=10)
        self.logger.addHandler(handler)
        self.logger.addHandler(logging.StreamHandler(sys.stderr))
        self.logger.setLevel(logging.DEBUG)
        formatter = logging.Formatter("%(name)s: %(levelname)s %(message)s")
        handler.setFormatter(formatter)

Internal = __Internal()


class IntDict(defaultdict, ReprMixin):

    """a dict where the values are expected to be numeric, so
    we can add dicts.If parent is given, parent is expected to
    be another IntDict, and our changes propagate into parent.
    This allows us to have a tree of IntDicts, and we only have
    to update the leaves, getting the sums for free"""

    def __init__(self, parent: Optional['IntDict']=None) ->None:
        super().__init__(int)
        self.parent = parent

    def copy(self) -> 'IntDict':
        """need to reimplement this because the __init__ signature of
        IntDict is not identical to that of defaultdict"""
        result = IntDict(self.parent)
        defaultdict.update(result, self)
        # see https://www.logilab.org/ticket/23986
        return result

    def __add__(self, other: object) -> 'IntDict':
        """add two IntDicts"""
        assert isinstance(other, IntDict), other
        result = self.copy()
        for key, value in other.items():
            result[key] += value
        return result

    def __radd__(self, other: object) -> 'IntDict':
        """we want sum to work (no start value)"""
        assert other == 0
        return self.copy()

    def __sub__(self, other: object) -> 'IntDict':
        """self - other"""
        assert isinstance(other, IntDict), other
        result = self.copy()
        for key, value in other.items():
            result[key] -= value
        for key in defaultdict.keys(result):
            if result[key] == 0:
                del result[key]
        return result

    def __eq__(self, other: object) ->bool:
        assert isinstance(other, IntDict), other
        return self.all() == other.all()

    def __ne__(self, other: object) ->bool:
        assert isinstance(other, IntDict), other
        return self.all() != other.all()

    def count(self, countFilter: Optional[Iterable]=None) ->int:
        """how many tiles defined by countFilter do we hold?
        countFilter is an iterator of element names. No countFilter: Take all
        So count(['we', 'ws']) should return 8"""
        return sum((defaultdict.get(self, x) or 0)  # type: ignore
                   for x in countFilter or self)

    def all(self, countFilter: Optional[Iterable]=None) ->List['Tile']:
        """return a list of all tiles defined by countFilter,
        each tile multiplied by its occurrence.
        countFilter is an iterator of element names. No countFilter: take all
        So all(['we', 'fs']) should return ['we', 'we', 'we', 'we', 'fs']"""
        result : List['Tile'] = []
        for element in countFilter or self:
            result.extend([element] * self[element])
        return sorted(result)

    def __contains__(self, tile: object) ->bool:
        """does not contain tiles with count 0"""
        return super().__contains__(tile) and self[tile] > 0

    def __setitem__(self, key: Any, value: Any) ->None:
        """also update parent if given"""
        if self.parent is not None:
            self.parent[key] += value - defaultdict.get(self, key, cast(Any, 0))
        super().__setitem__(key, value)

    def __delitem__(self, key: Any) ->None:
        """also update parent if given"""
        if self.parent is not None:
            self.parent[key] -= super().get(key, cast(Any, 0))
        super().__delitem__(key)

    def clear(self) ->None:
        """also update parent if given"""
        if self.parent is not None:
            for key, value in defaultdict.items(self):
                self.parent[key] -= value
        super().clear()

    def __str__(self) ->str:
        """sort the result for better log comparison"""
        keys = sorted(self.keys())
        return ', '.join(f'{str(x)}:{str(self[x])}' for x in keys)


class ZValues:

    """here we collect all zValues used in Kajongg"""
    itemZFactor = 100000
    boardZFactor = itemZFactor * 100
    markerZ = boardZFactor * 100 + 1
    movingZ = markerZ + 1
    popupZ = movingZ + 1


class Speeds:
    """some animation speeds in % of animationSpeed"""

    windDiscPercent = 70
    sideTextPercent = 80

    # The values below will change when animationSpeed changes.

    windDisc = 0
    sideText = 0

    @classmethod
    def refresh(cls, unusedOldValue:int, newValue:int) ->None:
        """recompute"""
        cls.windDisc = int(newValue * cls.windDiscPercent / 100)
        cls.sideText = int(newValue * cls.sideTextPercent / 100)


class DrawOnTopMixin:

    """The inheriting QGraphicsObject will draw itself above all non moving tiles"""

    def setDrawingOrder(self) ->None:
        """we want us above all non moving tiles"""
        if self.activeAnimation.get('pos'):  # type: ignore
            movingZ = ZValues.movingZ
        else:
            movingZ = 0
        self.setZValue(ZValues.markerZ + movingZ)  # type: ignore



# -*- coding: utf-8 -*-

"""
Copyright (C) 2008-2016 Wolfgang Rohdewald <wolfgang@rohdewald.de>

SPDX-License-Identifier: GPL-2.0-only

"""

from typing import Dict, TYPE_CHECKING, Any, Optional
from qt import QObject, QByteArray, QEvent, QSplitter, QHeaderView

from common import Internal, isAlive
from mi18n import english

if TYPE_CHECKING:
    from qt import QWidget


class StateSaver(QObject):

    """saves and restores the state for widgets"""

    savers : Dict['QWidget', 'StateSaver'] = {}

    def __init__(self, *widgets:'QWidget') ->None:
        super().__init__()
        pref = Internal.Preferences
        assert pref
        self.widgets = []
        for widget in widgets:
            if widget not in StateSaver.savers:
                StateSaver.savers[widget] = self
                widget.installEventFilter(self)
            name = self.__generateName(widget)
            self.widgets.append((name, widget))
            pref.addString('States', name + 'State')
            pref.addString('States', name + 'Geometry')
        for name, widget in self.widgets:
            stateFound = self.__restore(widget, name + 'State')
            geometryFound = self.__restore(widget, name + 'Geometry')
            if not stateFound and not geometryFound:
                pref.addString('States', name)
                self.__restore(widget, name)

    @staticmethod
    def __restore(widget:'QWidget', name:str) ->bool:
        """decode the saved string"""
        # pylint: disable=unsubscriptable-object
        def canRestore(name:str,what:str) ->bool:
            return name.endswith(what) and hasattr(widget, 'restore' + what)
        assert Internal.Preferences
        state = QByteArray.fromHex(Internal.Preferences[name].encode())
        if state:
            if canRestore(name, 'State'):
                widget.restoreState(state)  # type:ignore[attr-defined]
            elif canRestore(name, 'Geometry'):
                widget.restoreGeometry(state)
            else:
                # legacy
                if isinstance(widget, (QSplitter, QHeaderView)):
                    widget.restoreState(state)
                else:
                    widget.restoreGeometry(state)
        return bool(state)

    @staticmethod
    def __generateName(widget:'QWidget') ->str:
        """generate a name for this widget to be used in the config file"""
        _ = widget
        name = english(widget.objectName())
        if not name:
            while _.parentWidget():
                name = _.__class__.__name__ + name
                assert _  # for mypy - not possible because of while condition
                _ = _.parentWidget()
                if _.parentWidget():
                    widgetName = english(_.parentWidget().objectName())
                    if widgetName:
                        name = widgetName + name
                        break
        if not name:
            name = widget.__class__.__name__
        return str(name)

    def eventFilter(self, unusedWatched:Optional[QObject], event:Optional[QEvent]) ->bool:
        """if the watched widget hides, save its state.
        Return False if the event should be handled further"""
        if QEvent is not None and event:
            # while appquit, QEvent may be None. Maybe not anymore
            # with later versions?
            if event.type() in(QEvent.Type.Hide, QEvent.Type.Move, QEvent.Type.Resize):
                self.save()
            elif event.type() == QEvent.Type.Close:
                self.save()
                widget = self.widgets[0][1]
                if widget in StateSaver.savers:
                    del StateSaver.savers[widget]
        return False

    @staticmethod
    def saveAll() ->None:
        """execute all registered savers and write states to config file"""
        assert Internal.Preferences
        for saver in StateSaver.savers.values():
            saver.save()
        Internal.Preferences.writeConfig()

    @staticmethod
    def stateStr(state:Any) ->str:
        """convert hex string to str"""
        return str(bytes(state.toHex()).decode())

    def save(self) ->None:
        """writes the state into Preferences, but does not save"""
        assert Internal.Preferences
        for name, widget in self.widgets:
            if isAlive(widget):
                # pylint: disable=unsupported-assignment-operation
                if hasattr(widget, 'saveState'):
                    Internal.Preferences[name + 'State'] = self.stateStr(widget.saveState())
                if hasattr(widget, 'saveGeometry'):
                    Internal.Preferences[name + 'Geometry'] = self.stateStr(widget.saveGeometry())
        Internal.Preferences.writeConfig()



"""
Copyright (c) 2007-2008 Qtrac Ltd <mark@qtrac.eu>
Copyright (C) 2008-2016 Wolfgang Rohdewald <wolfgang@rohdewald.de>

SPDX-License-Identifier: GPL-2.0-only

"""

# mypy: disable_error_code="attr-defined,call-overload"
# Pyside6 does not seem to have complete/correct annotations for QStyleOptionViewItem

from typing import TYPE_CHECKING, Optional, Callable, Any, cast, Union

from qt import Qt, QSize, QRect, QEvent, QModelIndex
from qt import QStyledItemDelegate, QLabel, QTextDocument, QStyle, QPalette, \
    QStyleOptionViewItem, QApplication

from guiutil import Painter

if TYPE_CHECKING:
    from qt import QLocale, QObject, QPainter, QPersistentModelIndex, QAbstractItemModel
    from qt import QKeyEvent


class ZeroEmptyColumnDelegate(QStyledItemDelegate):

    """Display 0 or 0.00 as empty"""

    def displayText(self, value:Any, locale:'QLocale') ->str:  # type:ignore[override]
        """Display 0 or 0.00 as empty"""
        if isinstance(value, int) and value == 0:
            return ''
        if isinstance(value, float) and value == 0.0:
            return ''
        return super().displayText(value, locale)

class RichTextColumnDelegate(QStyledItemDelegate):

    """enables rich text in a view"""
    label:Optional[QLabel] = None
    document:QTextDocument

    def __init__(self, parent:Optional['QObject']=None) ->None:
        super().__init__(parent)
        if self.label is None:
            self.label = QLabel()
            self.label.setIndent(5)
            self.label.setTextFormat(Qt.TextFormat.RichText)
            self.document = QTextDocument()

    def paint(self, painter:Optional['QPainter'], option:QStyleOptionViewItem,
        index:Union[QModelIndex,'QPersistentModelIndex']) ->None:
        """paint richtext"""
        assert isinstance(index, QModelIndex)
        if option.state & QStyle.StateFlag.State_Selected:
            role = QPalette.ColorRole.Highlight
        else:
            role = QPalette.ColorRole.AlternateBase if index.row() % 2 else QPalette.ColorRole.Base
        assert self.label
        self.label.setBackgroundRole(role)
        if model := index.model():
            text = model.data(index, Qt.ItemDataRole.DisplayRole)
            self.label.setText(text)
        self.label.setFixedSize(option.rect.size())
        if painter:
            with Painter(painter):
                painter.translate(option.rect.topLeft())
                self.label.render(painter)

    def sizeHint(self, option:QStyleOptionViewItem, index:Union[QModelIndex,'QPersistentModelIndex']) ->QSize:
        """compute size for the final formatted richtext"""
        assert isinstance(index, QModelIndex)
        self.document.setDefaultFont(option.font)
        if model := index.model():
            self.document.setHtml(model.data(index))
        return QSize(int(self.document.idealWidth()) + 5,
                     option.fontMetrics.height())


class RightAlignedCheckboxDelegate(QStyledItemDelegate):

    """as the name says. From
https://wiki.qt.io/Technical_FAQ#How_can_I_align_the_checkboxes_in_a_view.3F"""

    def __init__(self, parent:'QObject', cellFilter:Callable) ->None:
        super().__init__(parent)
        self.cellFilter = cellFilter

    @staticmethod
    def __textMargin() ->int:
        """text margin"""
        if style := QApplication.style():
            return style.pixelMetric(
                QStyle.PixelMetric.PM_FocusFrameHMargin) + 1
        return 1

    def paint(self, painter:Optional['QPainter'], option:QStyleOptionViewItem,
        index:Union[QModelIndex,'QPersistentModelIndex']) ->None:
        """paint right aligned checkbox"""
        assert isinstance(index, QModelIndex)
        viewItemOption = QStyleOptionViewItem(option)
        if self.cellFilter(index):
            textMargin = self.__textMargin()
            newRect = QStyle.alignedRect(
                option.direction, Qt.AlignmentFlag.AlignRight,
                QSize(
                    option.decorationSize.width() + 5,
                    option.decorationSize.height()),
                QRect(
                    option.rect.x() + textMargin, option.rect.y(),
                    option.rect.width() - (2 * textMargin),
                    option.rect.height()))
            viewItemOption.rect = newRect
        if painter:
            super().paint(painter, viewItemOption, index)

    def editorEvent(self, event:Optional[QEvent], model:Optional['QAbstractItemModel'],
        option:QStyleOptionViewItem, index:Union[QModelIndex,'QPersistentModelIndex']) ->bool:
        """edit right aligned checkbox"""
        # pylint: disable=too-many-return-statements
        if not event or not model:
            return False
        assert isinstance(index, QModelIndex)
        flags = model.flags(index)
        # make sure that the item is checkable
        if not flags & Qt.ItemFlag.ItemIsUserCheckable or not flags & Qt.ItemFlag.ItemIsEnabled:
            return False
        # make sure that we have a check state
        value = index.data(Qt.ItemDataRole.CheckStateRole)
        if not isinstance(value, int):
            return False
        # make sure that we have the right event type
        if event.type() == QEvent.Type.MouseButtonRelease:
            textMargin = self.__textMargin()
            checkRect = QStyle.alignedRect(
                option.direction, Qt.AlignmentFlag.AlignRight,
                option.decorationSize,
                QRect(
                    option.rect.x() + (2 * textMargin), option.rect.y(),
                    option.rect.width() - (2 * textMargin),
                    option.rect.height()))
            if not checkRect.contains(event.pos()):  # type: ignore[attr-defined]
                return False
        elif event.type() == QEvent.Type.KeyPress:
            if cast('QKeyEvent', event).key() not in (Qt.Key.Key_Space, Qt.Key.Key_Select):
                return False
        else:
            return False
        if value == Qt.CheckState.Checked:
            state = Qt.CheckState.Unchecked
        else:
            state = Qt.CheckState.Checked
        return model.setData(index, state, Qt.ItemDataRole.CheckStateRole)



# -*- coding: utf-8 -*-

"""Copyright (C) 2013-2016 Wolfgang Rohdewald <wolfgang@rohdewald.de>

SPDX-License-Identifier: GPL-2.0-only

Read the user manual for a description of the interface to this scoring engine
"""

import itertools
from typing import List, Tuple, TYPE_CHECKING, Iterable, Dict, Any

from tile import Tile, Meld, MeldList

if TYPE_CHECKING:
    from tile import Tiles

class Permutations:

    """creates permutations for building melds out of single tiles.
    NEVER returns Kongs!"""
    cache : Dict[tuple, 'Permutations'] = {}
    permuteCache : Dict[tuple, Tuple[Tuple[Tuple, ...], ...]] = {}

    def __new__(cls, tiles:'Tiles') ->'Permutations':
        cacheKey = tuple(x.key for x in tiles)
        if cacheKey in cls.cache:
            return cls.cache[cacheKey]
        result = object.__new__(cls)
        cls.cache[cacheKey] = result
        return result

    def __init__(self, tiles:'Tiles') ->None:
        self.tiles = tiles
        if not hasattr(self, 'variants'):
            self.variants = self._variants()

    def _variants(self) ->List[MeldList]:
        """full Meld lists"""
        honors = []
        for tile in sorted(set(self.tiles)):
            if tile.isHonor:
                those = list(x for x in self.tiles if x == tile)
                count = len(those)
                if count == 4:
                    honors.append(Meld([those.pop()]))
                    count -= 1
                honors.append(Meld(those))
        boni = [x.single for x in self.tiles if x.isBonus]
        variants:List[List[MeldList]] = []
        for group in Tile.colors.upper():
            gTiles = [x for x in self.tiles if x.group == group]
            groupVariants = self.__colorVariants(group, [x.value for x in gTiles])  # type:ignore[misc]
            if groupVariants:
                variants.append(groupVariants)
        result:List[MeldList] = []
        variant:Any
        for variant in (sum(x, []) for x in itertools.product(*variants)):
            if variant not in result:
                result.append(MeldList(variant))
        result = sorted(MeldList(honors + x + boni) for x in result)
        return result

    @classmethod
    def permute(cls, valuesTuple:Tuple[int, ...]) ->Tuple[Tuple[Tuple[int, ...], ...], ...]:
        """return all groupings into melds.
        values is a tuple of int, range 1..9"""
        assert isinstance(valuesTuple, tuple)
        if valuesTuple in cls.permuteCache:
            return cls.permuteCache[valuesTuple]
        values = list(valuesTuple)
        result:List[Tuple[Tuple, ...]] = []
        possibleMelds = []
        valueSet = set(values)
        for value in sorted(valueSet):
            if values.count(value) == 2:
                possibleMelds.append(tuple([value] * 2))
            if values.count(value) >= 3:
                possibleMelds.append(tuple([value] * 3))
            if values.count(value + 1) and values.count(value + 2):
                possibleMelds.append(tuple([value, value + 1, value + 2]))
        if possibleMelds:
            for meld in possibleMelds:
                appendValue = list([meld])
                rest = values[:]
                for tile in meld:
                    rest.remove(tile)
                if rest:
                    permuteRest = cls.permute(tuple(rest))
                    for combi in permuteRest:
                        result.append(tuple(list(appendValue) + list(combi)))
                else:
                    result.append(tuple(appendValue))
        else:
            result = list([tuple([tuple([x]) for x in values])])  # pylint: disable=consider-using-generator
            # the solution proposed by pylint creates something unusable
        tupleResult = tuple(sorted(set(tuple(tuple(sorted(x)) for x in result))))
        cls.permuteCache[valuesTuple] = tupleResult
        return tupleResult

    colorPermCache : Dict[tuple, tuple] = {}

    @classmethod
    def usefulPermutations(cls, values:Iterable) ->Tuple:
        """return all variants usable for standard MJ formt (4 melds plus 1 pair),
        and also the variant with the most pungs. At least one will be returned.
        This is meant for the standard MJ format (4 pungs/kongs/chows plus 1 pair)"""
        values = tuple(values)
        if values not in cls.colorPermCache:
            variants = cls.permute(values)
            result = []
            maxPungs = -1
            maxPungVariant = minMeldVariant = None
            minMelds = 99
            for variant in variants:
                if all(len(meld) > 1 for meld in variant):
                    # no singles: usable for MJ
                    result.append(variant)
                if len(variant) < minMelds:
                    minMelds = len(variant)
                    minMeldVariant = variant
                pungCount = sum(
                    len(meld) == 3 and len(set(meld)) == 1 for meld in variant)
                if pungCount > maxPungs:
                    maxPungs = pungCount
                    maxPungVariant = variant
            if maxPungs > 0 and maxPungVariant not in result:
                assert maxPungVariant
                result.append(maxPungVariant)
            assert minMeldVariant
            result.append(minMeldVariant)
            if not result:
                # if nothing seems useful, return all possible permutations
                result.extend(variants)
            cls.colorPermCache[values] = tuple(result)
        return cls.colorPermCache[values]

    @classmethod
    def __colorVariants(cls, color:str, values:Iterable[int]) ->List[MeldList]:
        """generates all possible meld variants out of original
        where values is a string like '113445'.
        Returns lists of Meld"""
        allValues = sorted(values)
        vSet = set(allValues)
        groups = []
        for border in sorted(x + 1 for x in sorted(vSet) if x + 1 not in vSet):
            content = [x for x in allValues if x < border]
            if content:
                groups.append(content)
                allValues = [x for x in allValues if x > border]
        combinations = [cls.usefulPermutations(x) for x in groups]
        result:List[MeldList] = []
        for variant in itertools.product(*combinations):
            melds = MeldList()
            for block in variant:
                for meld in block:
                    melds.append(Meld(Tile(color, x) for x in meld))
            if melds:
                result.append(melds)
        return result



# -*- coding: utf-8 -*-

"""
Copyright (C) 2008-2016 Wolfgang Rohdewald <wolfgang@rohdewald.de>

SPDX-License-Identifier: GPL-2.0-only

"""

from typing import Any

from twisted.spread.pb import Error

from log import SERVERMARK


def srvMessage(*args: Any) ->str:
    """
    concatenate all args needed for i18n encoded in one string.
    For an explanation see log.translateServerMessage.

    @returns: The string to be wired.
    @rtype: C{str}, utf-8 encoded
    """
    strArgs = []
    for arg in args:
        if not isinstance(arg, str):
            arg = str(arg)
        strArgs.append(arg)
    return SERVERMARK + SERVERMARK.join(strArgs) + SERVERMARK


def srvError(*args: Any) ->'Error':
    """raise an exception, passing args as a single string"""
    return Error(srvMessage(*args))



# -*- coding: utf-8 -*-

"""
 Copyright (C) 2008-2016 Wolfgang Rohdewald <wolfgang@rohdewald.de>

SPDX-License-Identifier: GPL-2.0-only

"""

from typing import TYPE_CHECKING, Optional, Dict, Literal, Any, cast

from qt import Qt, QRectF, QPointF, QSizeF, QSize, QPen
from qt import QGraphicsObject, QGraphicsItem, QPixmap, QPainter, QColor, QFont

from util import stack
from log import logException, logDebug
from guiutil import Painter, sceneRotation
from common import LIGHTSOURCES, ZValues, Internal, Debug
from common import ReprMixin, isAlive, id4
from tile import Tile, Meld
from animation import AnimatedMixin

if TYPE_CHECKING:
    from qt import QKeyEvent, QWidget, QStyleOptionGraphicsItem
    from tileset import Tileset
    from board import Board


class UITile(AnimatedMixin, QGraphicsObject, ReprMixin):  # type:ignore[misc]

    """A tile visible on the screen. Every tile is only allocated once
    and then reshuffled and reused for every game.
    The unit of xoffset is the width of the tile,
    the unit of yoffset is the height of the tile.
    This is a QObject because we want to animate it."""


    clsUid = 0

    def __init__(self, tile:Tile, xoffset:float=0.0, yoffset:int=0, level:int=0) ->None:
        super().__init__()
        if not isinstance(tile, Tile):
            tile = Tile(tile)
        UITile.clsUid += 1
        self.uid = UITile.clsUid
        self._tile = tile
        self._boundingRect:Optional[QRectF] = None
        self._cross:bool = False
        self.setCacheMode(QGraphicsItem.CacheMode.DeviceCoordinateCache)
        # while moving the tile we use ItemCoordinateCache, see
        # Tile.setActiveAnimation
        self.__board:Optional['Board'] = None
        self.setClippingFlags()
        self.__xoffset = xoffset
        self.__yoffset = yoffset
        self.__dark = False
        self.level = level

    def debug_name(self) ->str:
        """identification for animations"""
        return self._tile.name2()

    def setClippingFlags(self) ->None:
        """if we do not show shadows, we need to clip"""
        assert Internal.Preferences
        showShadows = Internal.Preferences.showShadows
        self.setFlag(
            QGraphicsItem.GraphicsItemFlag.ItemClipsChildrenToShape,
            enabled=not showShadows)
        self.setFlag(QGraphicsItem.GraphicsItemFlag.ItemClipsToShape, enabled=not showShadows)

    def keyPressEvent(self, event:Optional['QKeyEvent']) ->None:
        """redirect to the board"""
        if event:
            _ = self.board
            assert _
            if self is not _.focusTile:
                logDebug(
                    f'keyPressEvent {event} on {self}_{id4(self)} but focus is on {_.focusTile}_{id4(_.focusTile)}')
            _.keyPressEvent(event)

    def __lightDistance(self) ->float:
        """the distance of item from the light source"""
        board = self.board
        if not board:
            return 0.0
        rect = self.sceneBoundingRect()
        lightSource = board.lightSource
        result = 0.0
        if 'E' in lightSource:
            result -= rect.right()
        if 'W' in lightSource:
            result += rect.left()
        if 'S' in lightSource:
            result -= rect.bottom()
        if 'N' in lightSource:
            result += rect.top()
        return result

    @property
    def tileset(self) ->Optional['Tileset']:
        """the active tileset"""
        return self.board.tileset if self.board else None

    def moveDict(self) ->Dict[Literal['pos', 'rotation', 'scale'], Any]:
        """a dict with attributes for the new position,
        normally pos, rotation and scale"""
        assert self.board
        assert self.tileset
        assert Internal.Preferences
        width = self.tileset.faceSize.width()
        height = self.tileset.faceSize.height()
        shiftZ = self.board.shiftZ(self.level)
        _ = self.yoffset
        if Internal.Preferences.showShadows and self.board.showShadowsBetweenRows:
            _ *= 1.2
        boardPos = QPointF(
            self.xoffset * width,
           _ * height)
        boardPos += shiftZ # QPointF + QPointF is unknown by given annotations
        scenePos = self.board.mapToScene(boardPos)
# TODO: rename to 'def get_moveDict, return class MoveDict(TypedDict)
        return {'pos': scenePos, 'rotation': sceneRotation(self.board), 'scale': self.board.scale()}

    def setDrawingOrder(self) ->None:
        """set drawing order for this tile"""
        if not isAlive(self):
            return
        if self.board:
            boardLevel = self.board.level
        else:
            boardLevel = ZValues.boardZFactor
        movingZ = 0
        # show moving tiles above non-moving tiles
        changePos = self.activeAnimation.get('pos')
        if changePos and not isAlive(changePos):
            return
        changeRotation = self.activeAnimation.get('rotation')
        if changeRotation and not isAlive(changeRotation):
            return
        changeScale = self.activeAnimation.get('scale')
        if changeScale and not isAlive(changeScale):
            return
        # show rotating and scaling tiles above all others
        if changeScale or changeRotation:
            movingZ += ZValues.movingZ
            movingZ += ZValues.boardZFactor
        elif changePos:
            if cast(float, self.rotation) % 180 == 0:
                currentY = self.y()
                newY = changePos.endValue().y()
            else:
                currentY = self.x()
                newY = changePos.endValue().x()
            if currentY != newY:
                movingZ += ZValues.movingZ
        self.setZValue(movingZ +
                       boardLevel +
                       (self.level + (2 if self.isKnown else 1))
                       * ZValues.itemZFactor +
                       self.__lightDistance())

    def boundingRect(self) ->QRectF:
        """define the part of the tile we want to see. Do not return QRect()
        if tileset is not known because that makes QGraphicsscene crash"""
        assert Internal.Preferences
        if self.tileset:
            self._boundingRect = QRectF(
                QPointF(),
                self.tileset.tileSize if Internal.Preferences.showShadows
                else self.tileset.faceSize)
        else:
            # just something. QRectF() gives segfault.
            self._boundingRect = QRectF(0.0, 0.0, 10.0, 10.0)
        return self._boundingRect

    def facePos(self, showShadows:Optional[bool]=None) ->QPointF:
        """return the face position relative to the tile
        depend on tileset, lightSource and shadow"""
        assert Internal.Preferences
        if showShadows is None:
            showShadows = bool(Internal.Preferences.showShadows)
        _ = self.board
        return _.tileFacePos(showShadows)

    def showFace(self) ->bool:
        """should we show face for this tile?"""
        return self.isKnown

    def __elementId(self, showShadows:Optional[bool]=None) ->str:
        """return the SVG element id of the tile"""
        assert Internal.Preferences
        if showShadows is None:
            showShadows = bool(Internal.Preferences.showShadows)
        if not showShadows:
            return "TILE_2"
        _ = self.board
        assert _
        lightSourceIndex = LIGHTSOURCES.index(_.rotatedLightSource())
        return f"TILE_{lightSourceIndex % 4 + 1}"

    def paint(self, painter:Optional[QPainter], unusedOption:Optional['QStyleOptionGraphicsItem'],
        unusedWidget:Optional['QWidget']=None) ->None:
        """paint the entire tile.
        I tried to cache a pixmap for the tile and darkener but without face,
        but that actually made it slower."""
        assert Internal.Preferences
        assert self.tileset
        assert painter
        with Painter(painter):
            renderer = self.tileset.renderer
            withBorders = Internal.Preferences.showShadows
            if not withBorders:
                painter.scale(*self.tileset.tileFaceRelation())
            renderer.render(painter, self.__elementId(), self.boundingRect())
            self._drawDarkness(painter)
        with Painter(painter):
            if self.showFace():
                if withBorders:
                    faceSize = self.tileset.faceSize.toSize()
                    renderer.render(
                        painter, self.svgName,
                        QRectF(self.facePos(), QSizeF(faceSize)))
                else:
                    renderer.render(
                        painter, self.svgName,
                        self.boundingRect())
        if self.cross:
            self.__paintCross(painter)
        if Debug.graphics or self.tile.exposed.name2() in Debug.focusable:
            self.__paintId(painter)

    def __paintId(self, painter:QPainter) ->None:
        """for debugging, paint uid"""
        with Painter(painter):
            assert self.tileset
            faceSize = self.tileset.faceSize
            height = faceSize.height()
            painter.translate(self.facePos())
            font = QFont()
            font.setPixelSize(24)
            painter.setFont(font)
            pen = QPen(QColor('orange'))
            painter.setPen(pen)
            painter.drawText(10, int(height / 2), str(self.uid))

    def __paintCross(self, painter:QPainter) ->None:
        """paint a cross on the tile"""
        with Painter(painter):
            assert self.tileset
            faceSize = self.tileset.faceSize
            width = faceSize.width()
            height = faceSize.height()
            painter.translate(self.facePos())
            painter.drawLine(QPointF(0.0, 0.0), QPointF(width, height))
            painter.drawLine(QPointF(width, 0.0), QPointF(0.0, height))

    def pixmapFromSvg(self, pmapSize:QSize, withBorders:Optional[bool]=None) ->QPixmap:
        """return a pixmap with default size as given in SVG
        and optional borders/shadows"""
        assert Internal.Preferences
        assert self.tileset
        if withBorders is None:
            withBorders = bool(Internal.Preferences.showShadows)
        if withBorders:
            originalSize = self.tileset.tileSize.toSize()
        else:
            originalSize = self.tileset.faceSize.toSize()
        result = QPixmap(pmapSize)
        result.fill(Qt.GlobalColor.transparent)
        painter = QPainter(result)
        if not painter.isActive():
            logException(
                f'painter is not active. Wanted size: {str(pmapSize)}')
        try:
            xScale = float(pmapSize.width()) / originalSize.width()
            yScale = float(pmapSize.height()) / originalSize.height()
        except ZeroDivisionError:
            xScale = 1
            yScale = 1
        # draw the tile too far to the left/upper side such that its shadow is outside of the print region
        if not withBorders:
            painter.scale(*self.tileset.tileFaceRelation())
        renderer = self.tileset.renderer
        renderer.render(painter, self.__elementId(showShadows=withBorders))
        painter.resetTransform()
        self._drawDarkness(painter)
        if self.showFace():
            faceSize = self.tileset.faceSize.toSize()
            faceSize = QSize(
                int(faceSize.width() * xScale),
                int(faceSize.height() * yScale))
            painter.resetTransform()
            painter.translate(self.facePos(withBorders))
            renderer.render(painter, self.tileset.svgName[self.exposed.name2()],
                            QRectF(QPointF(), QSizeF(faceSize)))
        return result

    def _drawDarkness(self, painter:QPainter) ->None:
        """if appropriate, make tiles darker. Mainly used for hidden tiles"""
        if self.dark:
            assert self.tileset
            board = self.board
            assert board
            rect = board.tileFaceRect().adjusted(-1, -1, -1, -1)
            color = QColor('black')
            color.setAlpha(self.tileset.darkenerAlpha)
            painter.fillRect(rect, color)

    def sortKey(self, sortDir:Qt.Key=Qt.Key.Key_Right) ->float:
        """moving order for cursor"""
        dirs = [Qt.Key.Key_Right, Qt.Key.Key_Up, Qt.Key.Key_Left, Qt.Key.Key_Down] * 2
        assert self.__board
        sorter = dirs[dirs.index(sortDir) + sceneRotation(self.__board) // 90]
        if sorter == Qt.Key.Key_Down:
            return self.xoffset * 100 + self.yoffset
        if sorter == Qt.Key.Key_Up:
            return -(self.xoffset * 100 + self.yoffset)
        if sorter == Qt.Key.Key_Left:
            return -(self.yoffset * 100 + self.xoffset)
        return self.yoffset * 100 + self.xoffset

    def setBoard(self, board:Optional['Board'],
        xoffset:Optional[float]=None, yoffset:Optional[int]=None, level:int=0) ->None:
        """change Position of tile in board"""
        placeDirty = False
        if self.__board != board:
            oldBoard = self.__board
            self.__board = board
            if oldBoard:
                oldBoard.removeUITile(self)
            if board:
                board.addUITile(self)
            placeDirty = True
        if self.level != level:
            self.level = level
            placeDirty = True
        if xoffset is not None and xoffset != self.__xoffset:
            self.__xoffset = xoffset
            placeDirty = True
        if yoffset is not None and yoffset != self.__yoffset:
            self.__yoffset = yoffset
            placeDirty = True
        if board and placeDirty:
            if board.scene() and not self.scene():
                if scene := board.scene():
                    scene.addItem(self)
            board.placeTile(self)

    @property
    def tile(self) ->Tile:
        """tile"""
        return self._tile

    @property
    def svgName(self) -> str:
        """shortcut"""
        assert self.tileset
        return self.tileset.svgName[self.exposed.name2()]

    def change_name(self, value:Tile) ->None:
        """set tile name and update display"""
        if value is not None:
            if self.name2() != value.name2():
                self._tile = value
                self.setDrawingOrder() # because known tiles are above unknown tiles
                self.update()

    @property
    def cross(self) ->bool:
        """cross tiles in kongbox"""
        return self._cross

    @cross.setter
    def cross(self, value:bool) ->None:
        """cross tiles in kongbox"""
        if self._cross == value:
            return
        self._cross = value
        self.update()

    @property
    def dark(self) ->bool:
        """show face?"""
        return self.__dark

    @dark.setter
    def dark(self, value:bool) ->None:
        """toggle and update display"""
        if value != self.__dark:
            self.__dark = value
            self.update()

    @property
    def focusable(self) ->bool:
        """as the name says"""
        return bool(self.flags() & QGraphicsItem.GraphicsItemFlag.ItemIsFocusable)

    @focusable.setter
    def focusable(self, value:bool) ->None:
        """redirect and generate Debug output"""
        if self.tile.name2() in Debug.focusable:
            newStr = 'focusable' if value else 'unfocusable'
            logDebug(f"{newStr}: {self} from {stack('')[-2]}")
        self.setFlag(QGraphicsItem.GraphicsItemFlag.ItemIsFocusable, value)

    @property
    def board(self) ->Optional['Board']:
        """get current board of this tile. Readonly."""
        return self.__board

    @property
    def xoffset(self) ->float:
        """in logical board coordinates"""
        return self.__xoffset

    @xoffset.setter
    def xoffset(self, value:float) ->None:
        """in logical board coordinates"""
        if value != self.__xoffset:
            self.__xoffset = value
            if self.__board:
                self.__board.placeTile(self)

    @property
    def yoffset(self) ->int:
        """in logical board coordinates"""
        return self.__yoffset

    @yoffset.setter
    def yoffset(self, value:int) ->None:
        """in logical board coordinates. Update board display."""
        if value != self.__yoffset:
            self.__yoffset = value
            if self.__board:
                self.__board.placeTile(self)

    def __str__(self) ->str:
        """printable string with tile"""
        if not Debug.graphics:
            return (f"{self.__class__.__name__}_{id4(self)}({self.tile.name2()} on "
                    f"{self.board.debug_name() if self.board else 'None'} "
                    f"x/y {self.xoffset:.1f}/{int(self.yoffset)})")
        rotation = f' rot{int(cast(float, self.rotation))}' if self.rotation else ''
        scale = f' scale={self.scale:.2f}' if self.scale != 1 else ''
        level = f' level={int(self.level)}' if self.level else ''
        _ = self.boundingRect()
        size = f' {int(_.width()):02}x{int(_.height()):02}'
        return (f"{self.__class__.__name__}_{id4(self)}({self.tile.name2()}"
                f" on {self.board.debug_name() if self.board else 'None'} "
                f" x/y/z {self.xoffset:.1f}/{self.x():.1f}/{self.yoffset:.1f} "
                f"{self.zValue():.1f}/{size}){rotation}{scale}{level}")

    @property
    def isBonus(self) ->bool:
        """proxy for tile"""
        return self.tile.isBonus

    @property
    def isKnown(self) ->bool:
        """proxy for tile"""
        return self.tile.isKnown

    @property
    def exposed(self) ->Tile:
        """proxy for tile"""
        return self.tile.exposed

    @property
    def concealed(self) ->Tile:
        """proxy for tile"""
        return self.tile.concealed

    @property
    def isConcealed(self) ->bool:
        """proxy for tile"""
        return self.tile.isConcealed

    @property
    def lowerGroup(self) ->str:
        """proxy for tile"""
        return self.tile.lowerGroup

    @property
    def char(self) ->str:
        """proxy for tile"""
        return self.tile.char

    def name2(self) ->str:
        """proxy for tile"""
        return self.tile.name2()


class UIMeld(list, ReprMixin):

    """represents a visible meld. Can be empty. Many Meld methods will
    raise exceptions if the meld is empty. But we do not care,
    those methods are not supposed to be called on empty melds.
    UIMeld is a list of UITile"""

    __hash__ = None  # type:ignore

    def __init__(self, newContent:Any) ->None:
        list.__init__(self)
        if (
                isinstance(newContent, list)
                and newContent
                and isinstance(newContent[0], UITile)):
            self.extend(newContent)
        elif isinstance(newContent, UITile):
            self.append(newContent)
        assert self, newContent

    @property
    def meld(self) ->Meld:
        """return a logical meld"""
        return Meld(x.tile for x in self)

    def __str__(self) ->str:
        """shorter than str() of the list"""
        first_tile = self[0]
        return (f'UIMeld_{id4(self)}({self.meld} in {first_tile.board.debug_name()} '
                f'x/y {first_tile.xoffset:.1f}/{int(first_tile.yoffset)})')



# -*- coding: utf-8 -*-

"""
Copyright (C) 2008-2016 Wolfgang Rohdewald <wolfgang@rohdewald.de>

SPDX-License-Identifier: GPL-2.0-only

"""

import logging
import os
from typing import TYPE_CHECKING, Union, Protocol, Optional

from locale import getpreferredencoding

# we must not import twisted or we need to change kajongg.py

from common import Internal, Debug, isAlive
from qt import Qt, QEvent, PYQT_VERSION
from util import elapsedSince, traceback, gitHead, callers
from mi18n import i18n
from dialogs import Sorry, Information, NoPrompt


if TYPE_CHECKING:
    from twisted.internet.defer import Deferred
    from twisted.python.failure import Failure
    from qt import QObject

SERVERMARK = '&&SERVER&&'


def translateServerMessage(msg:str) ->str:
    """because a PB exception can not pass a list of arguments, the server
    encodes them into one string using SERVERMARK as separator. That
    string is always english. Here we unpack and translate it into the
    client language."""
    if msg.find(SERVERMARK) >= 0:
        return i18n(*tuple(msg.split(SERVERMARK)[1:-1]))
    return msg

class SupportsIndent(Protocol):
    """for mypy"""
    indent:int

def dbgIndent(this:SupportsIndent, parent:Optional[SupportsIndent]) ->str:
    """show messages indented"""
    if this.indent == 0:
        return ''
    pIndent = parent.indent if parent and parent.indent else 0
    return (' │ ' * (pIndent)) + ' ├' + '─' * (this.indent - pIndent - 1)


def __logUnicodeMessage(prio:int, msg:str) ->None:
    """if we can encode the str msg to ascii, do so.
    Otherwise convert the str object into an utf-8 encoded
    str object.
    The logger module would log the str object with the
    marker feff at the beginning of every message, we do not want that."""
    byte_msg = msg.encode(getpreferredencoding(), 'ignore')[:4000]
    msg = byte_msg.decode(getpreferredencoding())
    Internal.logger.log(prio, msg)


def __enrichMessage(msg:str, withGamePrefix:bool=True) ->str:
    """
    Add some optional prefixes to msg: S/C, process id, time, git commit.

    @param msg: The original message.
    @type msg: C{str}
    @param withGamePrefix: If set, prepend the game prefix.
    @type withGamePrefix: C{Boolean}
    @rtype: C{str}
    """
    result = msg  # set the default
    if withGamePrefix and Internal.logPrefix:
        result = f"{Internal.logPrefix}{os.getpid() if Debug.process else ''}: {msg}"
    if Debug.timestamp:
        result = f'{elapsedSince(Debug.timestamp):08.4f} {result}'
    if Debug.git:
        head = gitHead()
        if head not in ('current', None):
            result = f'git:{head}/p3 {result}'
    if int(Debug.callers):
        result = '  ' + result
    return result


def __exceptionToString(exception:Exception) ->str:
    """
    Convert exception into a useful string for logging.

    @param exception: The exception to be logged.
    @type exception: C{Exception}

    @rtype: C{str}
    """
    parts = []
    for arg in exception.args:
        if hasattr(arg, 'strerror'):
            # when using py kde 4, this is already translated at this point
            # but I do not know what it does differently with gettext and if
            # I can do the same with the python gettext module
            parts.append(
                f'[Errno {arg.errno}] {i18n(arg.strerror)}')
        elif arg is None:
            pass
        else:
            parts.append(str(arg))
    if hasattr(exception, 'filename'):
        parts.append(exception.filename)
    return ' '.join(parts)

def logSummary(summary:traceback.StackSummary, prio:int) ->None:
    """log traceback summary"""
    for line in summary.format():
        if 'logException' not in line:
            __logUnicodeMessage(prio, '  ' + line.strip())

def logMessage(msg:Union[Exception, str], prio:int, showDialog:bool,
    showStack:bool=False, withGamePrefix:bool=True) ->'Deferred':
    """writes info message to log and to stdout"""
    if isinstance(msg, Exception):
        msg = __exceptionToString(msg)
    msg = str(msg)
    msg = translateServerMessage(msg)
    __logUnicodeMessage(prio, __enrichMessage(msg, withGamePrefix))
    if showStack:
        _ = traceback.walk_stack(None)
        summary = traceback.StackSummary.extract(_, limit=3, capture_locals=True)
        logSummary(summary, prio)
    if int(Debug.callers):
        __logUnicodeMessage(prio, '    ' + callers(int(Debug.callers)))
    if showDialog and not Internal.isServer:
        return Information(msg) if prio == logging.INFO else Sorry(msg, always=True)
    return NoPrompt(msg)


def logInfo(msg:Union[Exception, str], showDialog:bool=False, withGamePrefix:bool=True) ->'Deferred':
    """log an info message"""
    return logMessage(msg, logging.INFO, showDialog, withGamePrefix=withGamePrefix)


def logError(msg:Union[Exception, str], showStack:bool=True, withGamePrefix:bool=True) ->'Deferred':
    """log an error message"""
    return logMessage(msg, logging.ERROR, True, showStack=showStack, withGamePrefix=withGamePrefix)


def logDebug(msg:Union[Exception, str], showStack:bool=False,
    withGamePrefix:bool=True, btIndent:Optional[int]=None) ->'Deferred':
    """log this message and show it on stdout
    if btIndent is set, message is indented by depth(backtrace)-btIndent"""
    if btIndent:
        depth = traceback.extract_stack()
        msg = ' ' * (len(depth) - btIndent) + str(msg)
    return logMessage(msg, logging.DEBUG, False, showStack=showStack, withGamePrefix=withGamePrefix)


def logWarning(msg:Union[Exception, str], withGamePrefix:bool=True) ->'Deferred':
    """log this message and show it on stdout"""
    return logMessage(msg, logging.WARNING, True, withGamePrefix=withGamePrefix)

def logFailure(failure:'Failure') ->None:
    """twisted failure"""
    logError(f'{failure.getErrorMessage()} {failure.getTraceback()}')

def logException(exception: Union[Exception, str], withGamePrefix:bool=True) ->None:
    """logs error message and re-raises exception if we are not server"""
    logError(exception, withGamePrefix=withGamePrefix)
    exc_type = exception.__class__
    if hasattr(exception, 'type'):
        exc_type = exception.type
    isAssertion = 'Assertion' in exc_type.__name__
    if not Internal.isServer or isAssertion:
        if isinstance(exception, Exception):
            raise exception
        raise Exception(exception)  # pylint:disable=broad-exception-raised

class EventData(str):

    """used for generating a nice string"""
    events = {y: x for x, y in QEvent.__dict__.items() if isinstance(y, int)}
    # the following appear in Qt6/qtbase/src/corelib/kernel/qcoreevent.h but are not included.
    # if they were, we could just print event.type().name, but now name is only 15/16 etc for those
    extra = {
        15: 'Create',
        16: 'Destroy',
        152: 'AcceptDropsChange',
        154: 'Windows:ZeroTimer'
    }
    events.update(extra)
    keys = {y: x for x, y in Qt.__dict__.items() if isinstance(y, int)}

    prevFocusItem = None

    def __new__(cls, receiver:'QObject', event:'QEvent', prefix:Optional[str]=None) ->'EventData':
        """create the wanted string"""
        name = cls.eventName(event)
        msg = f'{name}.{cls.eventValue(event)}.receiver:{cls.eventReceiver(receiver)}'
        if prefix:
            msg = ': '.join([prefix, msg])
        if 'all' in Debug.events or any(x in msg for x in Debug.events.split(':')):
            logDebug(msg)
        return super().__new__(cls, msg)

    @classmethod
    def eventReceiver(cls, receiver:'QObject') ->str:
        """Format data about event receiver"""
        text = ''
        if hasattr(receiver, 'text'):
            if receiver.__class__.__name__ in ('QAbstractSpinBox', ):
                # this would segfault in text()
                text = ''
            else:
                try:
                    text = receiver.text()
                except TypeError:
                    text = receiver.text
        name = ''
        if hasattr(receiver, 'objectName') and receiver.objectName():
            name = receiver.objectName()
        debug_name = ''
        if hasattr(receiver, 'debug_name'):
            debug_name = receiver.debug_name()

        return ''.join([text, name, debug_name, repr(receiver)])

    @classmethod
    def eventValue(cls, event:'QEvent') ->str:
        """Format data about event value"""
        value = ''
        if hasattr(event, 'key'):
            if event.key() in cls.keys:
                value = cls.keys[event.key()]
            else:
                value = f'unknown key:{event.key()}'
        if hasattr(event, 'text'):
            eventText = str(event.text())
            if eventText and eventText != '\r':
                value += f':{eventText}'
        return value

    @classmethod
    def eventName(cls, event:'QEvent') ->str:
        """Format data about event name"""
        if not PYQT_VERSION:
            # Pyside
            evtype = event.type()
            repr_last = repr(evtype).split('.')[-1]
            result = repr_last.split(':')[0]
        elif event.type() in cls.events:
            # ignore unknown event types
            result = cls.events[event.type()]
        else:
            result = f'unknown :{event.type()}'
        return f'Event:{result}'

    def __init__(self, receiver, event, prefix=None):
        self.receiver = receiver
        self.event = event
        self.prefix = prefix
        if Debug.focusable and isAlive(Internal.scene):
            _ = Internal.scene.focusItem()
            if _ is not self.prevFocusItem:
                if EventData.prevFocusItem or _:
                    logDebug(f'{self} focusItem: {EventData.prevFocusItem} -> {_}')
                EventData.prevFocusItem = _

    def __str__(self):\
        return f'{self.eventName(self.event)} for {self.eventReceiver(self.receiver)}'
